from .base_creator import BaseCreator
from ...misc import utils

class gradclipperCreator(BaseCreator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    
    def prepare_packages(self):
        self.package = [utils]
    
    def maybe_modify_args(self, module_args):
        module_args['parameters'] = module_args['parameters'].parameters()
        return module_args


import numpy as np 
import torch

from .base_selector import BaseSelector
from ...config.setting.core_setting import (
    INDICES_TUPLE,
    INDICES_FLAG
)

class SemiHardSelector(BaseSelector):
    """
    Semi-hard sampling method, euclidean distance metric is required.
    """
    def __init__(self, margin=0.2, **kwargs):
        super(BaseSelector, self).__init__(**kwargs)
        self.margin = margin
    
    def forward(
        self,
        metric_mat,
        row_labels,
        col_labels,
        is_same_source=False
    ):
        """
        Randomly select a positive sample and select a negative sample holds:

        :math:`d_p < d_n < d_p + margin`
        """
        device = metric_mat.device
        bs = metric_mat.size(0)
        # pos and neg mask
        matches = (row_labels == col_labels).byte()
        diffs = matches ^ 1
        if is_same_source:
            matches.fill_diagonal_(0)

        has_pos_mask = torch.where(
            torch.sum(matches, dim=-1) > 0
        )[0]

        a_ids = torch.arange(bs)[has_pos_mask].to(device)
        # select positive samples
        p_ids = torch.multinomial(
            input=matches.float()[has_pos_mask, :],
            num_samples=1,
            replacement=True
        ).flatten()
        ap_dist = metric_mat[a_ids, p_ids].unsqueeze(1)

        # select negative samples
        has_pos_metric_mat = metric_mat[has_pos_mask]
        semi_hard_mask = (
            (has_pos_metric_mat > ap_dist) &
            (has_pos_metric_mat < self.margin + ap_dist)
        ).byte() * diffs[has_pos_mask]

        nonzero_semi_hard = torch.where(
            torch.sum(semi_hard_mask, dim=-1) > 0
        )[0] 
        n_ids = torch.multinomial(
            input=semi_hard_mask[nonzero_semi_hard, :].float(),
            num_samples=1,
            replacement=True
        ).flatten()
        a_ids = a_ids[nonzero_semi_hard]
        p_ids = p_ids[nonzero_semi_hard]

        indices_tuple = {
            INDICES_TUPLE: torch.stack([a_ids, p_ids, n_ids], dim=1),
            INDICES_FLAG: None
        }
        weight = None
        return (
            metric_mat,
            row_labels,
            col_labels,
            is_same_source,
            indices_tuple,
            weight,
        )


from torch.utils.data import Dataset
from abc import ABCMeta, abstractmethod

class BaseDataset(Dataset, metaclass=ABCMeta):
    """
    Base class of dataset. 

    Args:
        root (str):
            Root path of datasets
        phase (str):
            "train" or "test"
        assert_num_classes (int):
            Check whether the data is completely loaded.
        assert_num_samples (int):
            Check whether the data is completely loaded.
        transform (transform module):
            Image transform (augmentation). Default: None.
    """
    def __init__(self, root, phase, assert_num_classes, assert_num_samples, transform=None):
        self.root = root
        self.phase = phase
        self.assert_num_classes = assert_num_classes
        self.assert_num_samples = assert_num_samples
        self.transform = transform
        # initiate
        self.init_dataset()

    def set_transform(self, transform):
        self.transform = transform
    
    @abstractmethod
    def get_labels(self):
        pass

    @abstractmethod
    def init_dataset(self):
        pass

    @abstractmethod
    def __len__(self):
        pass

    def __getitem__(self, idx):
        img, label = self.get_image_label(idx)
        if self.transform is not None:
            img_dict = self.transform(img)
        else:
            raise NotImplementedError("No Transform!")
        info_dict = {
            "labels": label,
            "id": idx
        }
        info_dict.update(img_dict)
        return info_dict
    
    @abstractmethod
    def get_image_label(self, idx):
        pass
    


import torch
import numpy as np 
from torchvision import datasets
import os
import scipy.io as sio
import PIL.Image

from .base_dataset import BaseDataset

class Cars196(BaseDataset):
    """
    `Cars196 <https://ai.stanford.edu/~jkrause/cars/car_dataset.html>`_

    Use "cars_annos.mat" to load the images.
    """
    def init_dataset(self):
        self.root = os.path.join(self.root, "cars196")
        self.annos_file = os.path.join(self.root, "cars_annos.mat")
        annos_file = sio.loadmat(self.annos_file)

        if self.phase == "train":
            classes = range(0, 98)
        elif self.phase == "test":
            classes = range(98, 196)
        else:
            raise KeyError("Invalid self.phase!")
        
        paths = [
            item[0][0] for item 
            in annos_file["annotations"][0]
        ]
        labels = [
            int(item[5][0] - 1) for item 
            in annos_file["annotations"][0]
        ]

        self.paths, self.labels = [], []
        for path, label in zip(paths, labels):
            if label in classes:
                self.paths.append(os.path.join(self.root, path))
                self.labels.append(label)
        self.labels = np.array(self.labels)

        assert len(np.unique(self.labels)) == self.assert_num_classes
        assert self.__len__() == self.assert_num_samples

    def get_labels(self):
        return self.labels

    def __len__(self):
        return len(self.paths)
    
    def get_image_label(self, idx):
        # load image
        img = PIL.Image.open(self.paths[idx])
        # convert gray to rgb
        if len(list(img.split())) == 1:
            img = img.convert("RGB")
        # load label
        label = self.labels[idx]
        return img, label

# class Cars196(BaseDataset):
#     """
#     `Cars196 <https://ai.stanford.edu/~jkrause/cars/car_dataset.html>`_

#     NOTE: This old implementation has been depreciated!
#     """
#     def init_dataset(self):
#         self.root = os.path.join(self.root, "cars196")
#         img_folder = os.path.join(self.root, self.phase)
#         self.dataset = datasets.ImageFolder(img_folder)
#         self.labels = np.array([b for (a, b) in self.dataset.imgs])
#         assert len(np.unique(self.labels)) == self.assert_num_classes
#         assert self.__len__() == self.assert_num_samples

#     def get_labels(self):
#         return self.labels

#     def __len__(self):
#         return len(self.dataset)
    
#     def get_image_label(self, idx):
#         return self.dataset[idx]


import logging
import shutil
from sys import path
import pandas as pd 
import numpy as np 
import os

from gedml.launcher.misc import utils
from . import utils as utils_recorder

from .base_info_getter import BaseInfoGetter
from .base_info_writer import BaseInfoWriter

class BaseRecorder:
    """
    Recorder can search and save two forms of information:

    1. Attributions of module. If the module is inherited from ``WithRecorder`` or the module maintain a **to-record** list named 'to_record_list'.
    2. Dictionary with specific format:

    .. code:: python
    
        {
            table_name:
                item_name:
                    value
                    ...
        }

    Args:
        root (str):
            Path where to save the statistics and checkpoint.
        hint_if_exist (bool):
            Whether to hint if the folder is existed (in case that user delete the folder by mistake)
        delete_old_folder (bool):
            Whether to delete the old folder if the folder is existed.
    """
    def __init__(
        self,
        project_name="test",
        root="./",
        group_name="main",
        hint_if_exist=True,
        delete_old_folder=True,
        is_resume=False,
        use_wandb=True,
        params_to_save=None,
        link_config=None,
        pipeline_to_save=None,
        initiate_create=True,
    ):
        self.project_name = project_name
        self.root = root
        self.group_name = group_name
        self.exp_name = os.path.basename(self.root)
        self.hint_if_exist = hint_if_exist
        self.delete_old_folder = delete_old_folder
        self.is_resume = is_resume
        self.use_wandb = use_wandb
        self.params_to_save = params_to_save
        self.link_config = link_config
        self.pipeline_to_save = pipeline_to_save

        if initiate_create:
            self.create_exp_root()
            self.create_group_folders(group_name)

    def create_exp_root(self):
        self.root = utils_recorder.create_folder(
            path=self.root,
            is_resume=self.is_resume,
            hint_if_exist=self.hint_if_exist,
            delete_old_folders=self.delete_old_folder
        )

    def create_group_folders(self, group_name):
        sub_root_name = os.path.join(self.root, group_name)
        self.getter = BaseInfoGetter()
        self.writer = BaseInfoWriter(
            project_name=self.project_name,
            root=sub_root_name,
            exp_name=self.exp_name + "_{}".format(group_name),
            hint_if_exist=self.hint_if_exist,
            delete_old_folder=self.delete_old_folder,
            is_resume=self.is_resume,
            use_wandb=self.use_wandb,
            params_to_save=self.params_to_save,
            link_config=self.link_config,
            pipeline_to_save=self.pipeline_to_save
        )
    
    def delete_folders(self, group_name):
        sub_root_name = os.path.join(self.root, group_name)
        if os.path.exists(sub_root_name):
            shutil.rmtree(sub_root_name)
            logging.info("Delete group folder: {}".format(sub_root_name))

    
    def log_config(self, config_dict):
        self.writer.log_config(config_dict)
    
    def get_data(self, *args, **kwargs):
        """
        Extract the information from a module.

        Args:
            data (dict or obj):
                Dictionary of obj with **TO_RECORD_LIST** attribute.
            name (str):
                Specific name.
        
        Returns:
            tuple: (dict, int)

            1. output_dict (dict): information dictionary to be saved.
            2. step (int): step counted by ``recorder``.
        """
        return self.getter.get_data(*args, **kwargs)
    
    def update(self, *args, **kwargs):
        """
        Update record.

        Args:
            data (dict):
                Dictionary to be recorded.
            step (int):
                Step (optional).
        """
        self.writer.update(*args, **kwargs)
    
    def save_models(self, *args, **kwargs):
        """
        Save models.

        Args:
            obj (module):
                Modules with **TO_SAVE_LIST**.
            step (int):
                Step (or epoch)
            best (bool):
                Whether to be the best checkpoint.
            delete_old (bool):
                Whether to delete the old checkpoint.
        """
        self.writer.save_models(*args, **kwargs)
    
    def load_models(self, *args, **kwargs):
        """
        Load models.

        Args:
            obj (module):
                Modules with **TO_SAVE_LIST**.
            step (int):
                Step (or epoch).
            device (device):
                The computing device to load.
        """
        return self.writer.load_models(*args, **kwargs)
    
    def search_best_epoch(self, table_name, primary_key, maximum=True):
        """
        Search best epoch according given evaluation indicator.

        Args:
            table_name (str):
                The csv file.
            primary_key (str): 
                The primary_key to be searched.
            maximum (bool):
                Whether to regard the maximum to be the best.
        
        Returns:
            tuple: (best step, best metric)
        """
        table_path = os.path.join(self.writer.csv_path, table_name)
        df = pd.read_csv(table_path)
        # get primary metric
        selected_key = df[primary_key]
        if maximum:
            best_idx = selected_key.idxmax()
        else:
            best_idx = selected_key.idxmin()
        return df['step'][best_idx], df[primary_key][best_idx]
    


import torch
from abc import ABCMeta, abstractmethod

from ..modules import WithRecorder

class BaseSelector(WithRecorder, metaclass=ABCMeta):
    """
    Base class of ``selectors``.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    
    def forward(
        self, 
        metric_mat, 
        row_labels, 
        col_labels, 
        is_same_source=False
    ) -> tuple:
        """
        Args:
            metric_mat (torch.Tensor):
                Metric matrix.
            row_labels (torch.Tensor):
                Labels of rows.
            col_labels (torch.Tensor): 
                Labels of columns.
            is_same_source (bool):
                Whether the two data streams are from the same source.
        
        Returns:
            tuple: Five type of elements:

            1. metric_mat (torch.Tensor): Metric matrix.
            2. labels_row (torch.Tensor): Labels of rows.
            3. labels_col (torch.Tensor): Labels of columns.
            4. is_same_source (bool): Whether the two tensors are from the same source.
            5. indices_tuple (dict): Dict that has two key: "tuples" and "flags"
            6. weights (torch.Tensor): Weights.
        """
        indices_tuple, weights = None, None
        return (
            metric_mat, 
            row_labels, 
            col_labels, 
            is_same_source, 
            indices_tuple, 
            weights
        )



import torch

from ...misc import loss_function as l_f 
from ..base_loss import BaseLoss

class NPairLoss(BaseLoss):
    """
    Work with NPairSelector (Recommend)

    paper: `Improved deep metric learning with multi-class n-pair loss objective <http://www.nec-labs.com/uploads/images/Department-Images/MediaAnalytics/papers/nips16_npairmetriclearning.pdf>`_
    """
    def __init__(
        self,
        **kwargs
    ):
        super(NPairLoss, self).__init__(**kwargs)
    
    def required_metric(self):
        return ["dot_product"]
    
    def compute_loss(
        self,
        metric_mat,
        row_labels,
        col_labels,
        indices_tuple=None,
        weights=None,
        is_same_source=False,
    ) -> torch.Tensor:
        anchor_idx, positive_idx = l_f.get_unique_indices(indices_tuple, row_labels.squeeze())

        if len(anchor_idx) == 0:
            return torch.sum(metric_mat * 0)
        
        targets = torch.arange(len(anchor_idx)).to(metric_mat.device)
        sub_mat = metric_mat[anchor_idx, :][:, positive_idx]
        loss = torch.nn.functional.cross_entropy(sub_mat, targets)
        return loss


import torch
import logging
from tqdm import tqdm
import numpy as np 

from ..misc import utils

class BaseTester:
    """
    ``BaseTester`` takes charge of testing.

    Args:
        batch_size (int):
            Batch size.
        dataset_num_workers (int):
            Number of processes to load data.
        is_normalize (bool):
            Whether to normalize embeddings.
        splits_to_eval (list(str)):
            List of sub-dataset to evaluate.
        is_distributed (bool):
            Whether to set in the distributed mode.
    
    Example:
        >>> tester = BaseTester(32)
        >>> tester.prepare(...) # load models etc.
        >>> results = tester.test()
    """
    def __init__(
        self,
        batch_size,
        dataset_num_workers=8,
        is_normalize=True,
        splits_to_eval=["test"],
        is_distributed=False,
    ):
        self.batch_size = batch_size
        self.dataset_num_workers = dataset_num_workers
        self.is_normalize = is_normalize
        self.splits_to_eval = splits_to_eval
        self.is_distributed = is_distributed

        self.initiate_property()
    
    """
    Initialization
    """
    
    def initiate_property(self):
        self.trainable_object_list = [
            "models",
            "collectors",
        ]
        
    def initiate_datasets(self):
        self.datasets = {
            k: self.datasets[k] for k in self.splits_to_eval
        }
    
    """
    Set and Get
    """

    def set_batch_size(self, batch_size):
        self.batch_size = batch_size
    
    def set_distributed(self, flag=True):
        self.is_distributed = flag
    
    """
    test
    """
    def prepare(
        self,
        models,
        datasets,
        evaluators,
        device,
        device_ids,
    ):
        """
        Load objects to be tested.

        Args:
            models (dict):
                Dictionary of models.
            datasets (dict):
                Dictionary of datasets.
            evaluators (dict):
                Dictioanry of evaluators.
            device (device):
                Computation device.
            device_ids (list(int)):
                Instruct Faiss package to use the corresponding devices.
        """
        # pass parameters
        self.models = models
        self.datasets = datasets
        self.evaluators = evaluators
        self.device = device
        self.device_ids = device_ids

    def test(self):
        """
        Start testing.

        Returns:
            dict: evaluation results.
        """
        self.initiate_datasets()
        # start to test
        self.set_to_eval()
        outputs = {}
        with torch.no_grad():
            for k, v in self.datasets.items():
                # get the dataset loader
                self.initiate_dataloader(dataset=v)
                # get the embeddings
                self.get_embeddings()
                # compute the metrics
                results = self.compute_metrics()
                outputs[k] = results
        return outputs
    
    def set_to_eval(self):
        for trainable_name in self.trainable_object_list:
            trainable_object = getattr(self, trainable_name, None)
            if trainable_object  is None:
                logging.warn(
                    "{} is not a member of trainer".format(
                        trainable_name
                    )
                )
            else:
                for v in trainable_object.values():
                    v.eval()

    def initiate_dataloader(self, dataset):
        logging.info(
            "{}: Initiating dataloader".format(
                self.__class__.__name__
            )
        )
        sampler = None
        # get dataloader
        self.dataloader = torch.utils.data.DataLoader(
            dataset=dataset,
            batch_size=int(self.batch_size),
            sampler=sampler,
            drop_last=False,
            pin_memory=False,
            shuffle=False,
            num_workers=self.dataset_num_workers
        )
        self.dataloader_iter = iter(self.dataloader)
    
    def get_embeddings(self):
        logging.info(
            "Compute eval embeddings"
        )
        pbar = tqdm(self.dataloader_iter)
        embeddings_list, labels_list = [], []
        for info_dict in pbar:
            # get data
            data = info_dict["data"].to(self.device)
            label = info_dict["labels"].to(self.device)
            # forward
            embedding = self.compute_embeddings(data)
            embeddings_list.append(embedding)
            labels_list.append(label)
        self.embeddings = torch.cat(embeddings_list)
        self.labels = torch.cat(labels_list)

        # to numpy
        self.embeddings = self.embeddings.cpu().detach().numpy()
        self.labels = self.labels.cpu().numpy()
    
    def compute_embeddings(self, data):
        embedding = self.forward_models(data)
        return (
            torch.nn.functional.normalize(embedding, dim=-1) 
            if self.is_normalize 
            else embedding
        )
    
    def forward_models(self, data):
        embedding_trunk = self.models["trunk"](
            data
        )
        embedding_embedder = self.models["embedder"](
            embedding_trunk
        )
        return embedding_embedder
    
    def compute_metrics(self):
        metrics_dict = self.evaluators["default"].get_accuracy(
            self.embeddings,
            self.embeddings,
            self.labels,
            self.labels,
            True,
            device_ids=self.device_ids
        )
        return metrics_dict
    



import torch
import torch.nn.functional as F 
from . import (
    distance_metrics,
    similarity_metrics
)

class MetricFactory(torch.nn.Module):
    """
    Get different metric (distance or similarity)

    Args:
        is_normalize (bool):
            Whether to normalize the embeddings
        metric_name (str):
            'euclid', 'cosine', etc.
    
    Example:
        Get ``euclid`` metric:

        >>> metric = MetricFactory(is_normalize=False, metric_name="euclid")
        >>> data = torch.randn(100, 128)
        >>> matrix = metric(data, data)
        
    """
    def __init__(self, is_normalize, metric_name, addition=None, **kwargs):
        super().__init__(**kwargs)
        self.is_normalize = is_normalize
        self.metric_name = metric_name
        
        self.init_metric(addition)
    
    def init_metric(self, addition):
        self.metric_func = None
        for key in globals().keys():
            if "_metrics" in key:
                metric_module = globals()[key]
                self.metric_func = getattr(metric_module, self.metric_name, None)
                if self.metric_func is not None:
                    break
        assert self.metric_func is not None, "{} isn't valid! Please check 'metric_name'!"
        if addition is None:
            self.metric_func = self.metric_func()
        else:
            self.metric_func = self.metric_func(**addition)
        self.metric_type = self.metric_func.metric_type
        
    def forward(self, *args) -> torch.Tensor:
        """
        Get metric matrix.

        Args:
            *args (sequence):
                Sequence which is used to compute matrix.
        
        Returns:
            torch.Tensor: metric matrix.
        """
        args = list(args)
        if self.is_normalize:
            for idx, _ in enumerate(args):
                args[idx] = F.normalize(args[idx], dim=-1)
        metric_output = self.metric_func(*tuple(args))
        return metric_output
    

    


from .base_creator import BaseCreator
from ... import managers
from ...misc import utils
from ....config.setting.launcher_setting import (
    OBJECTS_KEY,
)

class managersCreator(BaseCreator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    
    def prepare_packages(self):
        self.package = [managers]



import torch

from ...misc import loss_function as l_f 
from ..base_loss import BaseLoss

class FastAPLoss(BaseLoss):
    """
    modified from: https://github.com/KevinMusgrave/pytorch-metric-learning

    paper: `Deep Metric Learning to Rank <https://openaccess.thecvf.com/content_CVPR_2019/html/Cakir_Deep_Metric_Learning_to_Rank_CVPR_2019_paper.html>`_
    """
    def __init__(
        self,
        num_bins=10,
        **kwargs
    ):
        super(FastAPLoss, self).__init__(**kwargs)
        self.num_bins = int(num_bins)
        self.num_edges = self.num_bins + 1
        self.histogram_max = 4
        self.histogram_delta = self.histogram_max / self.num_bins
    
    def required_metric(self):
        return ["euclid_normalized"]
    
    def compute_loss(
        self,
        metric_mat,
        row_labels,
        col_labels,
        indices_tuple=None,
        weights=None,
        is_same_source=False,
    ) -> torch.Tensor:
        dtype, device = metric_mat.dtype, metric_mat.device
        pos_mask = row_labels == col_labels
        neg_mask = ~ pos_mask
        if is_same_source:
            pos_mask.fill_diagonal_(False)
        pos_mask = pos_mask.type(dtype)
        neg_mask = neg_mask.type(dtype)

        N_pos = torch.sum(pos_mask, dim=1)
        safe_N = N_pos > 0
        if torch.sum(safe_N) == 0:
            return torch.sum(metric_mat * 0)
        
        # construct the histogram
        mid_points = (
            torch.linspace(0.0, self.histogram_max, steps=self.num_edges)
            .view(-1, 1, 1)
            .to(device)
            .type(dtype)
        )
        pulse = torch.relu(
            1 - torch.abs(metric_mat - mid_points) / self.histogram_delta
        )
        pos_hist = torch.sum(pulse * pos_mask, dim=2).t()
        neg_hist = torch.sum(pulse * neg_mask, dim=2).t()

        total_pos_hist = torch.cumsum(pos_hist, dim=1)
        total_hist = torch.cumsum(pos_hist + neg_hist, dim=1)

        h_pos_product = pos_hist * total_pos_hist
        safe_H = (h_pos_product > 0) & (total_hist > 0)
        if torch.sum(safe_H) > 0:
            FastAP = torch.zeros_like(pos_hist).to(device)
            FastAP[safe_H] = h_pos_product[safe_H] / total_hist[safe_H]
            FastAP = torch.sum(FastAP, dim=1)
            FastAP = FastAP[safe_N] / N_pos[safe_N]
            FastAP = 1 - FastAP
            return torch.mean(FastAP)
        else:
            return torch.sum(metric_mat * 0)


import os
import sys
import shutil
import logging
import torch

def get_first_key_of_dict(data: dict) -> str:
    first_key = list(data.keys())[0]
    return first_key

def create_folder(path, is_resume=False, hint_if_exist=True, delete_old_folders=True):
    """
    Create folder.

    Args:
        path (str):
            Path to be created.
        is_resume (bool):
            Whether to set resume mode. (default: False)
        hint_if_exist (bool):
            Hind if the folder is existed (It only works when delete_old_folders is True)
        delete_old_folders (bool):
            Whether to delete old folder. If it's True, old folder with the same name will be deleted.
            Otherwise, a "(x)" like suffix will be used to mark the new folder.
    
    Returns:
        str: Return the path if the path was modified.
    """
    if os.path.exists(path):
        if is_resume:
            logging.info("RESUME mode: not create new folder - {}!".format(path))
        else:
            if delete_old_folders:
                if hint_if_exist:
                    delete_choice = input("{path} exists! DELETE or NOT (Y/N): ".format(path=path))
                    if delete_choice != 'Y':
                        sys.exit(-1)
                shutil.rmtree(path)
                logging.info("{path} is RE-created!".format(path=path))
            else:
                ### use suffix to distinguish these folders with the same name.
                # check the current number of folders with the same name.
                suffix = "_({})"
                index = 2
                newpath = path + suffix.format(index)
                while(os.path.exists(newpath)):
                    index += 1
                    newpath = path + suffix.format(index)
                os.makedirs(newpath)
                path = newpath
                logging.info("New {path} is created!".format(path=path))
    else:
        os.makedirs(path)
        logging.info("New {path} is created!".format(path=path))
    
    return path

def meta_call_factory(obj, ordered_list, prefix, description_str=None, **kwargs):
    for func_suffix in ordered_list:
        func_name = prefix + func_suffix
        func_obj = getattr(obj, func_name)
        func_obj(**kwargs)
        if isinstance(description_str, str):
            logging.info(description_str.format(func_obj.__name__))

def get_value(data: (torch.Tensor or ...)):
    if isinstance(data, torch.Tensor):
        return data.item()
    elif isinstance(data, (int, float)):
        return data
    else:
        raise TypeError('Invalid data type: {}'.format(type(data)))



import torch
import math
from ..base_collector import BaseCollector

class ProxyCollector(BaseCollector):
    """
    Maintain proxy parameters to support proxy-based metric learning methods.

    Args:
        num_classes (int): Number of classes. default: 100.
        embeddings_dim (int): Dimension of embeddings. default: 128.
        centers_per_class (int): Number of centers per class. default: 1
    """
    def __init__(
        self,
        num_classes=100,
        embeddings_dim=128,
        centers_per_class=1,
        regularize_func="softtriple",
        regularize_weight=0,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.num_classes = num_classes
        self.embeddings_dim = embeddings_dim
        self.centers_per_class = centers_per_class
        self.regularize_func = regularize_func
        self.regularize_weight = regularize_weight

        self.initiate_regularize()
        self.initiate_params()

    def initiate_regularize(self):
        if isinstance(self.regularize_func, str):
            if self.regularize_func == "softtriple":
                self.register_buffer(
                    "pos_mask",
                    torch.zeros(self.num_classes*self.centers_per_class,
                    self.num_classes*self.centers_per_class, dtype=torch.bool)
                )
                K = self.centers_per_class
                for i in range(self.num_classes):
                    for j in range(K):
                        self.pos_mask[
                            i*K+j,
                            (i*K+j+1):(i*K+K)
                        ] = 1
                self.regularize_func = self._regularize_softtriple
            elif self.regularize_func == "structural":
                self.regularize_func = self._regularize_structural
    
    def _regularize_softtriple(self):
        proxies = torch.nn.functional.normalize(self.proxies, dim=-1, p=2)
        sim_mat_proxy = torch.matmul(proxies, proxies.t())
        reg_loss = torch.sum(
            torch.sqrt(2.0 + 1e-5 - 2. * sim_mat_proxy[self.pos_mask])
        ) / (self.num_classes * self.centers_per_class * (self.centers_per_class - 1))
        return reg_loss
    
    def _regularize_structural(self):
        proxies = torch.nn.functional.normalize(self.proxies, dim=-1, p=2)
        proxy_metric_mat = torch.matmul(proxies, proxies.t())
        proxy_pos_mask = (self.proxy_labels.unsqueeze(1) == self.proxy_labels.unsqueeze(0)).byte()
        proxy_neg_mask = proxy_pos_mask ^ 1
        proxy_pos_mask.fill_diagonal_(0)

        # compute pos metric
        pos_sum_exp = torch.sum(
            torch.exp(
                - proxy_metric_mat
            ) * proxy_pos_mask, dim=-1
        )
        neg_sum_exp = torch.sum(
            torch.exp(
                proxy_metric_mat
            ) * proxy_neg_mask, dim=-1
        )
        reg_loss = torch.mean(
            torch.log(
                1 + pos_sum_exp * neg_sum_exp
            )
        )
        return reg_loss

    
    def initiate_params(self):
        """
        Initiate proxies.
        """
        self.proxies = torch.nn.Parameter(
            torch.randn(
                self.num_classes * self.centers_per_class,
                self.embeddings_dim
            )
        )
        
        proxy_labels = (
            torch.arange(self.num_classes)
            .unsqueeze(1)
            .repeat(1, self.centers_per_class)
        ).flatten()
        self.register_buffer("proxy_labels", proxy_labels)
        torch.nn.init.kaiming_uniform_(self.proxies, a=math.sqrt(5))
    
    def forward(self, data, embeddings, labels) -> tuple:
        """
        Compute similarity (or distance) matrix between embeddings and proxies.
        """
        metric_mat = self.metric(embeddings, self.proxies)
        is_same_source = False

        # regularize multi-proxy
        if self.regularize_weight > 0 and self.centers_per_class > 1:
            reg_loss = self.regularize_func()
        else:
            reg_loss = 0
        return (
            metric_mat,
            labels.unsqueeze(-1),
            self.proxy_labels.unsqueeze(0),
            is_same_source,
            reg_loss,
            self.regularize_weight
        )



from .base_selector import BaseSelector

class DefaultSelector(BaseSelector):
    """
    Do nothing selector.
    """
    def forward(
        self, 
        metric_mat, 
        row_labels, 
        col_labels, 
        is_same_source=False
    ) -> tuple:
        """
        Do nothing.
        """
        indices_tuple, weight = None, None
        return metric_mat, row_labels, col_labels, is_same_source, indices_tuple, weight


import torch
import logging
import torch.nn.functional as F 
from abc import abstractmethod, ABCMeta

from ..modules import WithRecorder

class BaseLoss(WithRecorder, metaclass=ABCMeta):
    """
    Base loss module. The output of this module will be wrapped with "FINISH" flag which indicates the output doesn't need to be further processed.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    
    def forward(
        self,
        metric_mat,
        row_labels,
        col_labels,
        indices_tuple=None,
        weights=None,
        is_same_source=False,
    ) -> torch.Tensor:
        loss = self.compute_loss(
            metric_mat,
            row_labels,
            col_labels,
            indices_tuple=indices_tuple,
            weights=weights,
            is_same_source=is_same_source
        )
        return loss

    @abstractmethod
    def compute_loss(
        self,
        metric_mat,
        row_labels,
        col_labels,
        indices_tuple=None,
        weights=None,
        is_same_source=False,
    ) -> torch.Tensor:
        """
        Compute loss value.

        Args:
            metric_mat (torch.Tensor):
                Metric matrix.
            row_labels (torch.Tensor):
                Labels of matrix rows.
            col_labels (torch.Tensor):
                Labels of matrix columns.
            indices_tuple (dict):
                Dict that has two keys: "tuples" and "flags"
            weights (torch.Tensor):
                Can be element-wised, tuple-wised etc.
            is_same_source (bool):


        Returns:
            torch.Tensor: Final loss value (a tensor value).
        """
        return 0
    
    @abstractmethod
    def required_metric(self) -> list:
        return []


import torch

from .contrastive_loss import ContrastiveLoss

class SignalToNoiseRatioLoss(ContrastiveLoss):
    """
    paper: `Signal-To-Noise Ratio: A Robust Distance Metric for Deep Metric Learning <https://openaccess.thecvf.com/content_CVPR_2019/html/Yuan_Signal-To-Noise_Ratio_A_Robust_Distance_Metric_for_Deep_Metric_Learning_CVPR_2019_paper.html>`_
    """
    def __init__(
        self,
        **kwargs
    ):
        super(SignalToNoiseRatioLoss, self).__init__(**kwargs)
    
    def required_metric(self):
        return ["snr"]


import libtmux

def clear_tmux(session_name):
    """Kill the specified tmux session.

    Args:
        session_name(str):
            The name of tmux session to be killed.
    """
    target_session = TmuxManager(session_name=session_name)
    target_session.close_all()
    del target_session

class TmuxManager:
    """Class to manage the Tmux software.

    Note:
        Tmux must be installed before.

    Args:
        session_name(str):
            The name of tmux session.
        conda_env(str):
            The name of conda environment for window intialization.

    Example:
        >>> manager = TmuxManager()
        >>> manager.new_window(window_name="test")
        >>> cmd = "clear"
        >>> manager.send_cmd(cmd)
    
    Attributes:
        session:
            Current tmux session pointer.
        curr_window:
            Current tmux window pointer.
        curr_pane:
            Current tmux pane pointer by which user can send command.
    """
    def __init__(
        self,
        session_name='borel',
        conda_env="pytorch",
    ):
        self.session_name = session_name
        self.conda_env = conda_env

        self.initiate_manager()
    
    def initiate_manager(self):
        """Initiate libtmux.Server and create specified session.
        """
        self.server = libtmux.Server()
        # initiate session
        self.session = self.server.find_where(
            {
                "session_name": self.session_name
            }
        )
        if self.session is None:
            self.session = self.server.new_session(
                session_name=self.session_name
            )
        
        self.window_list = self.session.list_windows()
        self.curr_window = self.window_list[0]
        self.curr_pane = self.curr_window.list_panes()[0]
    
    def new_window(self, window_name=None, *args):
        """Create a new tmux window.

        After creating the tmux window, conda initialization command will be sent to enter into the specified conda environment.

        Args:
            window_name(str):
                The name of window to be created.
        """
        self.curr_window = self.session.new_window(
            window_name=window_name,
            attach=False,
        )
        self.window_list = self.session.list_windows()
        self.curr_pane = self.curr_window.list_panes()[0]
        # to conda environment
        cmd = "conda activate {}".format(self.conda_env)
        self.curr_pane.send_keys(cmd)
    
    def send_cmd(self, cmd):
        """Send command to the current tmux pane.

        Args:
            cmd(str):
                Command to be sent.
        """
        self.curr_pane.send_keys(cmd)
    
    def close_all(self):
        """Close all tmux window.
        """
        for w in self.window_list:
            w.kill_window()


    

    



import torch

from .base_distance_metric import BaseDistanceMetric

class LpMetric(BaseDistanceMetric):
    def __init__(self, p=2):
        self.p = p

    def __call__(self, row_input, col_input=None):
        if col_input is None:
            col_input = row_input
        return torch.cdist(row_input, col_input, p=self.p)



import torch
import functools

from ...config.setting.recorder_setting import TO_RECORD_LIST

class WithRecorder(torch.nn.Module):
    """
    Work with ``recorder protocol``.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        setattr(self, TO_RECORD_LIST, [])
    
    def add_recordable_attr(self, name: str):
        """
        Args:
            name (str):
                The name of attributes.
        """
        # add list
        if not getattr(self, TO_RECORD_LIST, None):
            setattr(self, TO_RECORD_LIST, [])
        # add attr
        to_record_list = getattr(self, TO_RECORD_LIST)
        if name not in to_record_list:
            to_record_list.append(name)
        if not getattr(self, name, None):
            setattr(self, name, 0)



from abc import ABCMeta, abstractmethod
from ....config.setting.launcher_setting import (
    CLASS_KEY,
    PARAMS_KEY,
    INITIATE_KEY,
)
import logging
import torch

class BaseCreator(metaclass=ABCMeta):
    def __init__(self, creator_mode=None):
        self.creator_mode = creator_mode
        self.init_packages()

    def __call__(self, module_params):
        assert isinstance(module_params, dict) 
        module_type = module_params[CLASS_KEY]
        module_args = module_params[PARAMS_KEY]
        self.creator_mode = module_params[INITIATE_KEY]
        module_args = {} if module_args is None else module_args
        return self.create(module_type, module_args)
    
    def init_packages(self):
        self.prepare_packages()
        assert isinstance(self.package, list)
    
    def register_packages(self, extra_package):
        """
        Register new packages.

        Args:
            extra_package (list or module):
                Extra packages to be registered.
        """
        if isinstance(extra_package, list):
            self.package = extra_package + self.package
        else:
            self.package = [extra_package] + self.package
        logging.info("{} has registered new packages {}".format(
            self.__class__.__name__, extra_package
        ))

    @abstractmethod
    def prepare_packages(self):
        self.package = None
    
    def create(self, module_type, module_args, *args, **kwargs):
        logging.info("+++ {}'s params: \n{}".format(
            module_type, self.print_params(module_args)
        ))
        # get class
        module_class = self.get_class(module_type)
        # create object
        return self.create_object(module_class, module_args)
    
    def print_params(self, args):
        args_string = ""
        for k, v in args.items():
            if isinstance(v, torch.nn.Module):
                v = "module-" + v.__class__.__name__
            elif isinstance(v, dict):
                v = "dict-" + ";".join(v.keys())
            args_string += "\t\t\t{}: {}\n".format(k, v)
        return args_string
    
    def get_class(self, module_type):
        module_class = None
        for sub_package in self.package:
            module_class = getattr(sub_package, module_type, None)
            if module_class is not None:
                break
        assert module_class is not None, "Error Info: {} - {}".format(
            sub_package, module_type
        )
        return module_class
    
    def create_object(self, module_class, module_args):
        module_args = self.maybe_modify_args(module_args)
        module_object = module_class(**module_args)
        module_object = self.maybe_modify_object(module_object)
        return module_object
    
    def maybe_modify_args(self, module_args):
        return module_args

    def maybe_modify_object(self, module_object):
        return module_object


import numpy as np 
import torch

from .base_selector import BaseSelector
from ...config.setting.core_setting import (
    INDICES_TUPLE,
    INDICES_FLAG
)

class DistanceWeightedSelector(BaseSelector):
    """
    Distance weighted sampling method, euclidean distance metric is required.

    paper: `Sampling Matters in Deep Embedding Learning <http://openaccess.thecvf.com/content_iccv_2017/html/Wu_Sampling_Matters_in_ICCV_2017_paper.html>`_
    """
    def __init__(self, lower_cutoff=0.5, upper_cutoff=1.4, embedding_dim=512, **kwargs):
        super(DistanceWeightedSelector, self).__init__(**kwargs)
        self.lower_cutoff = lower_cutoff
        self.upper_cutoff = upper_cutoff
        self.embedding_dim = embedding_dim
    
    def forward(
        self,
        metric_mat,
        row_labels,
        col_labels,
        is_same_source=False
    ) -> tuple:
        """
        Randomly select a positive sample for anchor sample and select a negative sample for anchor sample according to the distance weighted probability.

        The distribution of pairwise distances follows:

        :math:`q(d) \propto d^{n-2} [1 - \\frac{1}{4} d^2 ]^{\\frac{n-3}{2}}`
        """
        metric_mat_to_weights = metric_mat.clamp(min=self.lower_cutoff).detach()
        bs = metric_mat.size(0)
        device = metric_mat.device
        # pos and neg mask
        matches = (row_labels == col_labels).byte()
        diffs = matches ^ 1

        # compute distance weighted prob
        minus_log_weights = (
            (2.0 - self.embedding_dim) * torch.log(metric_mat_to_weights) + 
            ((3.0 - self.embedding_dim) / 2.0) * torch.log(1.0 - 0.25 * metric_mat_to_weights.pow(2))
        ) * diffs
        ## avoid inf or nan
        inf_or_nan = torch.isinf(minus_log_weights) | torch.isnan(minus_log_weights)
        ## compute weight
        inv_weights = torch.exp(
            minus_log_weights - torch.max(minus_log_weights[~inf_or_nan])
        )
        ## NOTE: Cutting of values with high distances made the results slightly worse according to Confusezius / Pytorch-Metric-Learning-Baseline
        # inv_weights = (
        #     inv_weights * diffs * (metric_mat_to_weights < self.upper_cutoff).byte()
        # )
        inv_weights[inf_or_nan] = 0

        # construct the tuple
        if is_same_source:
            matches.fill_diagonal_(0)

        # avoid nan, inf and zero
        non_zero_row = torch.where(
            (torch.sum(matches, dim=-1) > 0) & (torch.sum(inv_weights, dim=-1) > 0)
        )[0]
        inv_weights = inv_weights[non_zero_row, :]
        ## anchor indices
        a_ids = torch.arange(bs)[non_zero_row].to(device)
        ## positive indices
        p_ids = torch.multinomial(
            input=matches[non_zero_row, :].float(),
            num_samples=1,
            replacement=True
        ).flatten()
        ## negative indices
        inv_weights = inv_weights / inv_weights.sum(dim=-1, keepdim=True)
        n_ids = torch.multinomial(
            input=inv_weights,
            num_samples=1,
            replacement=True
        ).flatten()

        indices_tuple = {
            INDICES_TUPLE: torch.stack([a_ids, p_ids, n_ids], dim=1),
            INDICES_FLAG: None
        }
        weight = None
        return (
            metric_mat,
            row_labels,
            col_labels,
            is_same_source,
            indices_tuple,
            weight,
        )


import torch

from ...misc import loss_function as l_f 
from ..base_loss import BaseLoss

class ContrastiveLoss(BaseLoss):
    """
    paper: `Learning a Similarity Metric Discriminatively, with Application to Face Verification <https://ieeexplore.ieee.org/abstract/document/1467314/>`_
    """
    def __init__(
        self,
        pos_margin=0,
        neg_margin=1,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.pos_margin = pos_margin
        self.neg_margin = neg_margin

        to_record_list = [
            "mean_pos_pair_dist",
            "mean_neg_pair_dist",
            "mean_pos_loss",
            "mean_neg_loss",
            "nonzero_mean_pos_loss",
            "nonzero_mean_neg_loss"
        ]
        for item in to_record_list:
            self.add_recordable_attr(name=item)
    
    def required_metric(self):
        return ["euclid"]
    
    def compute_loss(
        self,
        metric_mat,
        row_labels,
        col_labels,
        indices_tuple,
        is_same_source=False,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        pos_pair, neg_pair = l_f.indices_to_pairs(metric_mat, indices_tuple)

        self.mean_pos_pair_dist = torch.mean(pos_pair)
        self.mean_neg_pair_dist = torch.mean(neg_pair)

        pos_loss = torch.nn.functional.relu(pos_pair - self.pos_margin)
        neg_loss = torch.nn.functional.relu(self.neg_margin - neg_pair)

        self.mean_pos_loss = torch.mean(pos_loss)
        self.mean_neg_loss = torch.mean(neg_loss)

        pos_loss = pos_loss[torch.where(pos_loss)[0]]
        neg_loss = neg_loss[torch.where(neg_loss)[0]]

        self.nonzero_mean_pos_loss = torch.mean(pos_loss)
        self.nonzero_mean_neg_loss = torch.mean(neg_loss)

        numerator = torch.sum(pos_loss) + torch.sum(neg_loss)
        denominator = len(pos_loss) + len(neg_loss)
        loss = numerator / denominator
        return loss
    



# from .metrics import *
import logging
from . import metrics
import numpy as np 

class CalculatorFromMat:
    """
    A evaluation manager class. Given ``query`` and ``referecne`` vectors, this module will return a evaluation dictionary including all evaluation results (Given metric matrix)

    Args:
        k_list (list):
            K-index list for ``recall at k`` and ``precision at k``.
        include (tuple):
            Evaluation functions must be included.
        exclude (tuple):
            Evaluation functions should be excluded.
        k (int):
            Parameter for knn search.
        is_min (bool):
            Whether to choose the min (or max) value as the "nearest" sample.

    Example:
        >>> calculator = CalculatorFromMat(k_list=[1,2,4,8])
        >>> x = np.random.randn(100, 128)
        >>> labels = np.random.randint(0, 10, size=(100,))
        >>> output_dict = calculator.get_accuracy(x, x, labels, labels, True)
    """
    def __init__(self, k_list, include=(), exclude=(), k=None, is_min=True):
        assert isinstance(k_list, list)
        self.k_list = k_list
        self.k = k
        self.function_keyword = "calculate_"
        self.meta_function_keyword = "_meta_calculate_"
        self.include = include
        self.exclude = exclude
        self.is_min = is_min

        self.init_original_function_dict()
        self.check_primary_metrics(include, exclude)
        self.original_function_dict = self.get_function_dict(include, exclude)
        self.curr_function_dict = self.get_function_dict()
    
    def check_primary_metrics(self, include=(), exclude=()):
        primary_metrics = list(self.original_function_dict.keys())
        for met in [include, exclude]:
            if not isinstance(met, (tuple, list)):
                raise TypeError("Arguments must be of type tuple, not {}.".format(type(met)))
            if not set(met).issubset(set(primary_metrics)):
                raise ValueError("Primary metrics must be one or more of: {}.".format(primary_metrics))

    def meta_metrics(self):
        return ['precision_at_k', 'recall_at_k']

    def requires_knn(self): 
        return ["mean_average_precision_at_r", "r_precision"]
    
    def init_original_function_dict(self):
        # prepare other metrics except recall@k and precision@k
        function_names = [x for x in dir(self) if x.startswith(self.function_keyword)]
        metrics = [x.replace(self.function_keyword, "", 1) for x in function_names]
        self.original_function_dict = {x:getattr(self, y) for x, y in zip(metrics, function_names)}

        # get recall and precision function
        for meta_metric in self.meta_metrics():
            meta_metric_names, meta_metric_functions = [], []
            for k in self.k_list:
                meta_metric_names.append(meta_metric.replace('at_k', "at_"+str(k)))
                meta_metric_functions.append(getattr(self, self.meta_function_keyword+meta_metric)(k))
            self.original_function_dict.update(zip(meta_metric_names, meta_metric_functions))
    
    def get_function_dict(self, include=(), exclude=()):
        if len(include) == 0:
            include = list(self.original_function_dict.keys())
        include_metrics = [k for k in include if k not in exclude]
        return {k:v for k, v in self.original_function_dict.items() if k in include_metrics}
    
    def get_curr_metrics(self):
        return [k for k in self.curr_function_dict.keys()]
    
    def _meta_calculate_precision_at_k(self, k):
        def calculate_precision_at_k(knn_labels, query_labels, not_lone_query_mask, **kwargs):
            if not any(not_lone_query_mask):
                return 0
            knn_labels, query_labels = knn_labels[not_lone_query_mask], query_labels[not_lone_query_mask]
            return metrics.precision_at_k(knn_labels, query_labels[:, None], k)
        return calculate_precision_at_k
    
    def _meta_calculate_recall_at_k(self, k):
        def calculate_recall_at_k(knn_labels, query_labels, not_lone_query_mask, **kwargs):
            if not any(not_lone_query_mask):
                return 0
            knn_labels, query_labels = knn_labels[not_lone_query_mask], query_labels[not_lone_query_mask]
            return metrics.recall_at_k(knn_labels, query_labels, k)
        return calculate_recall_at_k
    
    def calculate_mean_average_precision_at_r(self, knn_labels, query_labels, not_lone_query_mask, embeds_same_source, label_counts, **kwargs):
        if not any(not_lone_query_mask):
            return 0
        knn_labels, query_labels = knn_labels[not_lone_query_mask], query_labels[not_lone_query_mask]
        return metrics.mean_average_precision_at_r(knn_labels, query_labels[:, None], embeds_same_source, label_counts)
    
    def calculate_r_precision(self, knn_labels, query_labels, not_lone_query_mask, embeds_same_source, label_counts, **kwargs):
        if not any(not_lone_query_mask):
            return 0
        knn_labels, query_labels = knn_labels[not_lone_query_mask], query_labels[not_lone_query_mask]
        return metrics.r_precision(knn_labels, query_labels[:, None], embeds_same_source, label_counts)
    
    def get_accuracy(self, metric_mat: np.ndarray, query_labels: np.ndarray, reference_labels: np.ndarray, embeds_same_source: bool, include=(), exclude=(), device_ids=None):
        """
        Compute all evaluation indicators. 

        Args:
            metric_mat (np.ndarray):
                Distance matrix to be tested. size: :math:`B_1 \\times B_2`.
            query_labels (np.ndarray):
                Queris' labels. size: :math:`B_1`.
            reference_labels (np.ndarray):
                References' labels. size: :math:`B_2`.
            embeds_same_source (bool):
                Whether ``query`` and ``reference`` data are from the same source.
            include (tuple):
                Evaluation functions to include.
            exclude (tuple):
                Evaluation functions to exclude.
            device_ids (list): 
                Device indices to call by Faiss package. default: None.
            
        Returns:
            dict: A dictionary which contains all results.
        """
        logging.info('Start computing metrics!...')
        self.curr_function_dict = self.get_function_dict(include, exclude)

        kwargs = {
            "metric_mat": metric_mat,
            "query_labels": query_labels,
            "reference_labels": reference_labels,
            "embeds_same_source": embeds_same_source
        }

        if any(x in self.requires_knn() for x in self.get_curr_metrics()):
            label_counts, num_k = metrics.get_label_counts(reference_labels)
            if self.k is not None: num_k = self.k
            assert num_k >= max(self.k_list) # max(k) mustn't samller than the max element of k_list
            knn_indices, knn_distances = metrics.get_knn_from_mat(metric_mat, num_k, embeds_same_source, is_min=self.is_min, device_ids=device_ids)
            knn_labels = reference_labels[knn_indices]
            lone_query_labels = metrics.get_lone_query_labels(query_labels, reference_labels, label_counts, embeds_same_source)
            not_lone_query_mask = ~np.isin(query_labels, lone_query_labels)

            if not any(not_lone_query_mask):
                logging.warning("None of the query labels are in the reference set.")
            kwargs["label_counts"] = label_counts
            kwargs["knn_labels"] = knn_labels
            kwargs["knn_distances"] = knn_distances
            kwargs["lone_query_labels"] = lone_query_labels
            kwargs["not_lone_query_mask"] = not_lone_query_mask
            kwargs["device_ids"] = device_ids
        
        return self._get_accuracy(self.curr_function_dict, **kwargs)
    
    def _get_accuracy(self, function_dict, **kwargs):
        return {k:v(**kwargs) for k,v in function_dict.items()}
                



import torch

from ...misc import loss_function as l_f 
from ..base_loss import BaseLoss

class ProxyAnchorLoss(BaseLoss):
    """
    paper: `Proxy Anchor Loss for Deep Metric Learning <http://openaccess.thecvf.com/content_CVPR_2020/html/Kim_Proxy_Anchor_Loss_for_Deep_Metric_Learning_CVPR_2020_paper.html>`_
    """
    def __init__(
        self,
        margin=0.1,
        alpha=32,
        **kwargs
    ):
        super(ProxyAnchorLoss, self).__init__(**kwargs)
        self.alpha = alpha
        self.margin = margin
    
    def required_metric(self):
        return ["cosine"]
    
    def compute_loss(
        self,
        metric_mat,
        row_labels,
        col_labels,
        indices_tuple=None,
        weights=None,
        is_same_source=False,
    ) -> torch.Tensor:
        pos_mask = row_labels == col_labels
        neg_mask = ~ pos_mask

        with_pos_proxies = torch.where(torch.sum(pos_mask, dim=0) != 0)[0]

        pos_term = l_f.logsumexp(
            - self.alpha * (metric_mat - self.margin),
            keep_mask=pos_mask,
            add_one=True,
            dim=0
        ).squeeze()
        neg_term = l_f.logsumexp(
            self.alpha * (metric_mat + self.margin),
            keep_mask=neg_mask,
            add_one=True,
            dim=0
        ).squeeze()
        
        if len(with_pos_proxies) == 0:
            pos_loss = torch.sum(metric_mat * 0)
        else:
            pos_loss = torch.mean(pos_term[with_pos_proxies])
        neg_loss = torch.mean(neg_term)

        return pos_loss + neg_loss


import numpy as np 
import torch
import logging
from scipy.special import comb
# sklearn
from sklearn.cluster import KMeans
from sklearn.neighbors import NearestNeighbors
from sklearn.decomposition import PCA
# faiss
try:
    import faiss
except ModuleNotFoundError:
    logging.warning("Faiss Package Not Found! (metrics package)")

from ..misc import utils

# cluster algorithm
def get_knn(ref_embeds, embeds, k, embeds_same_source=False, device_ids=None):
    d = ref_embeds.shape[1]
    if device_ids is not None:
        index = faiss.IndexFlatL2(d)
        index = utils.index_cpu_to_gpu_multiple(index, gpu_ids=device_ids)
        index.add(ref_embeds)
        distances, indices = index.search(embeds, k+1)
        if embeds_same_source:
            return indices[:, 1:], distances[:, 1:]
        else:
            return indices[:, :k], distances[:, :k]
    else:
        neigh = NearestNeighbors(n_neighbors=k)
        neigh.fit(ref_embeds)
        distances, indices = neigh.kneighbors(embeds, k + 1)
        if embeds_same_source:
            return indices[:, 1:], distances[:, 1:]
        else:
            return indices[:, :k], distances[:, :k]

def get_knn_from_mat(metric_mat, k, embeds_same_source=False, is_min=True, device_ids=None):
    device = torch.device("cuda:{}".format(device_ids[0]))
    metric_mat = torch.from_numpy(metric_mat).to(device)
    # sort
    sorted_value, sorted_indices = torch.sort(metric_mat, dim=-1, descending=not is_min)
    if embeds_same_source:
        return (
            sorted_indices[:, 1:(k+1)].cpu().numpy(),
            sorted_value[:, 1:(k+1)].cpu().numpy()
        )
    else:
        return (
            sorted_indices[:, :k].cpu().numpy(),
            sorted_value[:, :k].cpu().numpy()
        )

def run_kmeans(x, num_clusters, device_ids=None):
    _, d = x.shape
    if device_ids is not None:
        # faiss implementation of k-means
        clus = faiss.Clustering(d, num_clusters)
        clus.niter = 20
        clus.max_points_per_centroid = 10000000
        index = faiss.IndexFlatL2(d)
        index = utils.index_cpu_to_gpu_multiple(index, gpu_ids=device_ids)
        # perform the training
        clus.train(x, index)
        _, idxs = index.search(x, 1)
        return np.array([int(n[0]) for n in idxs], dtype=np.int64)
    else:
        # k-means
        kmeans = KMeans(n_clusters=num_clusters, random_state=0).fit(x)
        return kmeans.labels_ 

def run_pca(x, out_dim, device_ids=None):
    if device_ids is not None:
        mat = faiss.PCAMatrix(x.shape[1], out_dim)
        mat.train(x)
        assert mat.is_trained
        return mat.apply_py(x)
    else:
        pca = PCA(n_components=out_dim)
        data_output = pca.fit_transform(x)
        return data_output

# metrics functions: code from: github: pytorch-metric-learning
def get_relevance_mask(shape, gt_labels, embeds_same_source, label_counts):
    relevance_mask = np.zeros(shape=shape, dtype=np.int)
    for k, v in label_counts.items():
        matching_rows = np.where(gt_labels==k)[0]
        max_column = v-1 if embeds_same_source else v
        relevance_mask[matching_rows, :max_column] = 1
    return relevance_mask

def get_label_counts(ref_labels):
    unique_labels, label_counts = np.unique(ref_labels, return_counts=True)
    num_k = min(1023, int(np.max(label_counts)))
    return {k:v for k, v in zip(unique_labels, label_counts)}, num_k

def get_lone_query_labels(query_labels, ref_labels, ref_label_counts, embeds_same_source):
    if embeds_same_source:
        return np.array([k for k, v in ref_label_counts.items() if v <= 1])
    else:
        return np.setdiff1d(query_labels, ref_labels)

def r_precision(knn_labels, gt_labels, embeds_same_source, label_counts):
    relevance_mask = get_relevance_mask(knn_labels.shape, gt_labels, embeds_same_source, label_counts)
    matches_per_row = np.sum((knn_labels == gt_labels) * relevance_mask.astype(bool), axis=1)
    max_possible_matches_per_row = np.sum(relevance_mask, axis=1)
    accuracy_per_sample = matches_per_row / max_possible_matches_per_row
    return np.mean(accuracy_per_sample)

def mean_average_precision_at_r(knn_labels, gt_labels, embeds_same_source, label_counts):
    relevance_mask = get_relevance_mask(knn_labels.shape, gt_labels, embeds_same_source, label_counts)
    num_samples, num_k = knn_labels.shape
    equality = (knn_labels == gt_labels) * relevance_mask.astype(bool)
    cumulative_correct = np.cumsum(equality, axis=1)
    k_idx = np.tile(np.arange(1, num_k+1), (num_samples, 1))
    precision_at_ks = (cumulative_correct * equality) / k_idx
    summed_precision_pre_row = np.sum(precision_at_ks * relevance_mask, axis=1)
    max_possible_matches_per_row = np.sum(relevance_mask, axis=1)
    accuracy_per_sample = summed_precision_pre_row / max_possible_matches_per_row
    return np.mean(accuracy_per_sample)

def precision_at_k(knn_labels, gt_labels, k):
    curr_knn_labels = knn_labels[:, :k]
    accuracy_per_sample = np.sum(curr_knn_labels == gt_labels, axis=1) / k 
    return np.mean(accuracy_per_sample)

def recall_at_k(knn_labels, gt_labels, k):
    accuracy_per_sample = np.array([float(gt_label in recalled_predictions[:k]) for gt_label, recalled_predictions in zip(gt_labels, knn_labels)])
    return np.mean(accuracy_per_sample)

def f1_score(query_labels, cluster_labels):
    # compute tp_plus_fp
    qlabels_set, qlabels_counts = np.unique(query_labels, return_counts=True)
    tp_plut_fp = sum([comb(item, 2) for item in qlabels_counts if item > 1])

    # compute tp
    tp = sum([sum([comb(item, 2) for item in np.unique(cluster_labels[query_labels==query_label], return_counts=True)[1] if item > 1]) for query_label in qlabels_set])

    # compute fp
    fp = tp_plut_fp - tp

    # compute fn
    fn = sum([comb(item, 2) for item in np.unique(cluster_labels, return_counts=True)[1] if item > 1]) - tp

    # compute F1
    P, R = tp / (tp+fp), tp / (tp+fn)
    F1 = 2*P*R / (P+R)
    return F1


import torch
import numpy as np 
from pandas import read_csv
import os
from PIL import Image
from .base_dataset import BaseDataset

class OnlineProducts(BaseDataset):
    """
    `Stanford Online Products <https://cvgl.stanford.edu/projects/lifted_struct/>`_
    """
    def init_dataset(self):
        self.root = os.path.join(self.root, "online_products")
        info_folder = os.path.join(self.root, 'Info_Files')
        img_folder = os.path.join(self.root, 'images')
        self.img_paths, self.labels = [], []
        self.init_info_file_name()

        curr_file = read_csv(
            os.path.join(info_folder, self.info_file_name),
            delim_whitespace=True, 
            header=0
        ).values
        self.img_paths.extend(
            [os.path.join(img_folder, name) 
            for name in list(curr_file[:, 3])]
        )
        self.labels.extend(list(curr_file[:, 1] - 1))
        
        self.labels = np.array(self.labels)
        assert len(np.unique(self.labels)) == self.assert_num_classes
        assert self.__len__() == self.assert_num_samples
    
    def init_info_file_name(self):
        if self.phase == "train":
            self.info_file_name = "Ebay_train.txt"
        elif self.phase == "test":
            self.info_file_name = "Ebay_test.txt"
        else:
            raise KeyError(
                "Invalid dataset phase: {} / {}".format(
                    self.phase,
                    self.__class__.__name__
                )
            )

    def get_labels(self):
        return self.labels
    
    def __len__(self):
        return len(self.labels)

    def get_image_label(self, idx):
        path = self.img_paths[idx]
        img = Image.open(path).convert("RGB")
        label = self.labels[idx]
        return img, label


import torch
import torch.nn as nn
import torch.utils.model_zoo as model_zoo
import os
import sys
import types

__all__ = ['BNInception', 'bninception']

pretrained_settings = {
    'bninception': {
        'imagenet': {
            # Was ported using python2 (may trigger warning)
            'url': 'http://data.lip6.fr/cadene/pretrainedmodels/bn_inception-52deb4733.pth',
            # 'url': 'http://yjxiong.me/others/bn_inception-9f5701afb96c8044.pth',
            'input_space': 'BGR',
            'input_size': [3, 224, 224],
            'input_range': [0, 255],
            'mean': [104, 117, 128],
            'std': [1, 1, 1],
            'num_classes': 1000
        }
    }
}

class BNInception(nn.Module):

    def __init__(self, num_classes=1000):
        super(BNInception, self).__init__()
        inplace = True
        self.conv1_7x7_s2 = nn.Conv2d(3, 64, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3))
        self.conv1_7x7_s2_bn = nn.BatchNorm2d(64, affine=True)
        self.conv1_relu_7x7 = nn.ReLU (inplace)
        self.pool1_3x3_s2 = nn.MaxPool2d ((3, 3), stride=(2, 2), dilation=(1, 1), ceil_mode=True)
        self.conv2_3x3_reduce = nn.Conv2d(64, 64, kernel_size=(1, 1), stride=(1, 1))
        self.conv2_3x3_reduce_bn = nn.BatchNorm2d(64, affine=True)
        self.conv2_relu_3x3_reduce = nn.ReLU (inplace)
        self.conv2_3x3 = nn.Conv2d(64, 192, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.conv2_3x3_bn = nn.BatchNorm2d(192, affine=True)
        self.conv2_relu_3x3 = nn.ReLU (inplace)
        self.pool2_3x3_s2 = nn.MaxPool2d ((3, 3), stride=(2, 2), dilation=(1, 1), ceil_mode=True)
        self.inception_3a_1x1 = nn.Conv2d(192, 64, kernel_size=(1, 1), stride=(1, 1))
        self.inception_3a_1x1_bn = nn.BatchNorm2d(64, affine=True)
        self.inception_3a_relu_1x1 = nn.ReLU (inplace)
        self.inception_3a_3x3_reduce = nn.Conv2d(192, 64, kernel_size=(1, 1), stride=(1, 1))
        self.inception_3a_3x3_reduce_bn = nn.BatchNorm2d(64, affine=True)
        self.inception_3a_relu_3x3_reduce = nn.ReLU (inplace)
        self.inception_3a_3x3 = nn.Conv2d(64, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_3a_3x3_bn = nn.BatchNorm2d(64, affine=True)
        self.inception_3a_relu_3x3 = nn.ReLU (inplace)
        self.inception_3a_double_3x3_reduce = nn.Conv2d(192, 64, kernel_size=(1, 1), stride=(1, 1))
        self.inception_3a_double_3x3_reduce_bn = nn.BatchNorm2d(64, affine=True)
        self.inception_3a_relu_double_3x3_reduce = nn.ReLU (inplace)
        self.inception_3a_double_3x3_1 = nn.Conv2d(64, 96, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_3a_double_3x3_1_bn = nn.BatchNorm2d(96, affine=True)
        self.inception_3a_relu_double_3x3_1 = nn.ReLU (inplace)
        self.inception_3a_double_3x3_2 = nn.Conv2d(96, 96, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_3a_double_3x3_2_bn = nn.BatchNorm2d(96, affine=True)
        self.inception_3a_relu_double_3x3_2 = nn.ReLU (inplace)
        self.inception_3a_pool = nn.AvgPool2d (3, stride=1, padding=1, ceil_mode=True, count_include_pad=True)
        self.inception_3a_pool_proj = nn.Conv2d(192, 32, kernel_size=(1, 1), stride=(1, 1))
        self.inception_3a_pool_proj_bn = nn.BatchNorm2d(32, affine=True)
        self.inception_3a_relu_pool_proj = nn.ReLU (inplace)
        self.inception_3b_1x1 = nn.Conv2d(256, 64, kernel_size=(1, 1), stride=(1, 1))
        self.inception_3b_1x1_bn = nn.BatchNorm2d(64, affine=True)
        self.inception_3b_relu_1x1 = nn.ReLU (inplace)
        self.inception_3b_3x3_reduce = nn.Conv2d(256, 64, kernel_size=(1, 1), stride=(1, 1))
        self.inception_3b_3x3_reduce_bn = nn.BatchNorm2d(64, affine=True)
        self.inception_3b_relu_3x3_reduce = nn.ReLU (inplace)
        self.inception_3b_3x3 = nn.Conv2d(64, 96, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_3b_3x3_bn = nn.BatchNorm2d(96, affine=True)
        self.inception_3b_relu_3x3 = nn.ReLU (inplace)
        self.inception_3b_double_3x3_reduce = nn.Conv2d(256, 64, kernel_size=(1, 1), stride=(1, 1))
        self.inception_3b_double_3x3_reduce_bn = nn.BatchNorm2d(64, affine=True)
        self.inception_3b_relu_double_3x3_reduce = nn.ReLU (inplace)
        self.inception_3b_double_3x3_1 = nn.Conv2d(64, 96, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_3b_double_3x3_1_bn = nn.BatchNorm2d(96, affine=True)
        self.inception_3b_relu_double_3x3_1 = nn.ReLU (inplace)
        self.inception_3b_double_3x3_2 = nn.Conv2d(96, 96, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_3b_double_3x3_2_bn = nn.BatchNorm2d(96, affine=True)
        self.inception_3b_relu_double_3x3_2 = nn.ReLU (inplace)
        self.inception_3b_pool = nn.AvgPool2d (3, stride=1, padding=1, ceil_mode=True, count_include_pad=True)
        self.inception_3b_pool_proj = nn.Conv2d(256, 64, kernel_size=(1, 1), stride=(1, 1))
        self.inception_3b_pool_proj_bn = nn.BatchNorm2d(64, affine=True)
        self.inception_3b_relu_pool_proj = nn.ReLU (inplace)
        self.inception_3c_3x3_reduce = nn.Conv2d(320, 128, kernel_size=(1, 1), stride=(1, 1))
        self.inception_3c_3x3_reduce_bn = nn.BatchNorm2d(128, affine=True)
        self.inception_3c_relu_3x3_reduce = nn.ReLU (inplace)
        self.inception_3c_3x3 = nn.Conv2d(128, 160, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))
        self.inception_3c_3x3_bn = nn.BatchNorm2d(160, affine=True)
        self.inception_3c_relu_3x3 = nn.ReLU (inplace)
        self.inception_3c_double_3x3_reduce = nn.Conv2d(320, 64, kernel_size=(1, 1), stride=(1, 1))
        self.inception_3c_double_3x3_reduce_bn = nn.BatchNorm2d(64, affine=True)
        self.inception_3c_relu_double_3x3_reduce = nn.ReLU (inplace)
        self.inception_3c_double_3x3_1 = nn.Conv2d(64, 96, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_3c_double_3x3_1_bn = nn.BatchNorm2d(96, affine=True)
        self.inception_3c_relu_double_3x3_1 = nn.ReLU (inplace)
        self.inception_3c_double_3x3_2 = nn.Conv2d(96, 96, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))
        self.inception_3c_double_3x3_2_bn = nn.BatchNorm2d(96, affine=True)
        self.inception_3c_relu_double_3x3_2 = nn.ReLU (inplace)
        self.inception_3c_pool = nn.MaxPool2d ((3, 3), stride=(2, 2), dilation=(1, 1), ceil_mode=True)
        self.inception_4a_1x1 = nn.Conv2d(576, 224, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4a_1x1_bn = nn.BatchNorm2d(224, affine=True)
        self.inception_4a_relu_1x1 = nn.ReLU (inplace)
        self.inception_4a_3x3_reduce = nn.Conv2d(576, 64, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4a_3x3_reduce_bn = nn.BatchNorm2d(64, affine=True)
        self.inception_4a_relu_3x3_reduce = nn.ReLU (inplace)
        self.inception_4a_3x3 = nn.Conv2d(64, 96, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_4a_3x3_bn = nn.BatchNorm2d(96, affine=True)
        self.inception_4a_relu_3x3 = nn.ReLU (inplace)
        self.inception_4a_double_3x3_reduce = nn.Conv2d(576, 96, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4a_double_3x3_reduce_bn = nn.BatchNorm2d(96, affine=True)
        self.inception_4a_relu_double_3x3_reduce = nn.ReLU (inplace)
        self.inception_4a_double_3x3_1 = nn.Conv2d(96, 128, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_4a_double_3x3_1_bn = nn.BatchNorm2d(128, affine=True)
        self.inception_4a_relu_double_3x3_1 = nn.ReLU (inplace)
        self.inception_4a_double_3x3_2 = nn.Conv2d(128, 128, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_4a_double_3x3_2_bn = nn.BatchNorm2d(128, affine=True)
        self.inception_4a_relu_double_3x3_2 = nn.ReLU (inplace)
        self.inception_4a_pool = nn.AvgPool2d (3, stride=1, padding=1, ceil_mode=True, count_include_pad=True)
        self.inception_4a_pool_proj = nn.Conv2d(576, 128, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4a_pool_proj_bn = nn.BatchNorm2d(128, affine=True)
        self.inception_4a_relu_pool_proj = nn.ReLU (inplace)
        self.inception_4b_1x1 = nn.Conv2d(576, 192, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4b_1x1_bn = nn.BatchNorm2d(192, affine=True)
        self.inception_4b_relu_1x1 = nn.ReLU (inplace)
        self.inception_4b_3x3_reduce = nn.Conv2d(576, 96, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4b_3x3_reduce_bn = nn.BatchNorm2d(96, affine=True)
        self.inception_4b_relu_3x3_reduce = nn.ReLU (inplace)
        self.inception_4b_3x3 = nn.Conv2d(96, 128, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_4b_3x3_bn = nn.BatchNorm2d(128, affine=True)
        self.inception_4b_relu_3x3 = nn.ReLU (inplace)
        self.inception_4b_double_3x3_reduce = nn.Conv2d(576, 96, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4b_double_3x3_reduce_bn = nn.BatchNorm2d(96, affine=True)
        self.inception_4b_relu_double_3x3_reduce = nn.ReLU (inplace)
        self.inception_4b_double_3x3_1 = nn.Conv2d(96, 128, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_4b_double_3x3_1_bn = nn.BatchNorm2d(128, affine=True)
        self.inception_4b_relu_double_3x3_1 = nn.ReLU (inplace)
        self.inception_4b_double_3x3_2 = nn.Conv2d(128, 128, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_4b_double_3x3_2_bn = nn.BatchNorm2d(128, affine=True)
        self.inception_4b_relu_double_3x3_2 = nn.ReLU (inplace)
        self.inception_4b_pool = nn.AvgPool2d (3, stride=1, padding=1, ceil_mode=True, count_include_pad=True)
        self.inception_4b_pool_proj = nn.Conv2d(576, 128, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4b_pool_proj_bn = nn.BatchNorm2d(128, affine=True)
        self.inception_4b_relu_pool_proj = nn.ReLU (inplace)
        self.inception_4c_1x1 = nn.Conv2d(576, 160, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4c_1x1_bn = nn.BatchNorm2d(160, affine=True)
        self.inception_4c_relu_1x1 = nn.ReLU (inplace)
        self.inception_4c_3x3_reduce = nn.Conv2d(576, 128, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4c_3x3_reduce_bn = nn.BatchNorm2d(128, affine=True)
        self.inception_4c_relu_3x3_reduce = nn.ReLU (inplace)
        self.inception_4c_3x3 = nn.Conv2d(128, 160, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_4c_3x3_bn = nn.BatchNorm2d(160, affine=True)
        self.inception_4c_relu_3x3 = nn.ReLU (inplace)
        self.inception_4c_double_3x3_reduce = nn.Conv2d(576, 128, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4c_double_3x3_reduce_bn = nn.BatchNorm2d(128, affine=True)
        self.inception_4c_relu_double_3x3_reduce = nn.ReLU (inplace)
        self.inception_4c_double_3x3_1 = nn.Conv2d(128, 160, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_4c_double_3x3_1_bn = nn.BatchNorm2d(160, affine=True)
        self.inception_4c_relu_double_3x3_1 = nn.ReLU (inplace)
        self.inception_4c_double_3x3_2 = nn.Conv2d(160, 160, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_4c_double_3x3_2_bn = nn.BatchNorm2d(160, affine=True)
        self.inception_4c_relu_double_3x3_2 = nn.ReLU (inplace)
        self.inception_4c_pool = nn.AvgPool2d (3, stride=1, padding=1, ceil_mode=True, count_include_pad=True)
        self.inception_4c_pool_proj = nn.Conv2d(576, 128, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4c_pool_proj_bn = nn.BatchNorm2d(128, affine=True)
        self.inception_4c_relu_pool_proj = nn.ReLU (inplace)
        self.inception_4d_1x1 = nn.Conv2d(608, 96, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4d_1x1_bn = nn.BatchNorm2d(96, affine=True)
        self.inception_4d_relu_1x1 = nn.ReLU (inplace)
        self.inception_4d_3x3_reduce = nn.Conv2d(608, 128, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4d_3x3_reduce_bn = nn.BatchNorm2d(128, affine=True)
        self.inception_4d_relu_3x3_reduce = nn.ReLU (inplace)
        self.inception_4d_3x3 = nn.Conv2d(128, 192, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_4d_3x3_bn = nn.BatchNorm2d(192, affine=True)
        self.inception_4d_relu_3x3 = nn.ReLU (inplace)
        self.inception_4d_double_3x3_reduce = nn.Conv2d(608, 160, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4d_double_3x3_reduce_bn = nn.BatchNorm2d(160, affine=True)
        self.inception_4d_relu_double_3x3_reduce = nn.ReLU (inplace)
        self.inception_4d_double_3x3_1 = nn.Conv2d(160, 192, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_4d_double_3x3_1_bn = nn.BatchNorm2d(192, affine=True)
        self.inception_4d_relu_double_3x3_1 = nn.ReLU (inplace)
        self.inception_4d_double_3x3_2 = nn.Conv2d(192, 192, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_4d_double_3x3_2_bn = nn.BatchNorm2d(192, affine=True)
        self.inception_4d_relu_double_3x3_2 = nn.ReLU (inplace)
        self.inception_4d_pool = nn.AvgPool2d (3, stride=1, padding=1, ceil_mode=True, count_include_pad=True)
        self.inception_4d_pool_proj = nn.Conv2d(608, 128, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4d_pool_proj_bn = nn.BatchNorm2d(128, affine=True)
        self.inception_4d_relu_pool_proj = nn.ReLU (inplace)
        self.inception_4e_3x3_reduce = nn.Conv2d(608, 128, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4e_3x3_reduce_bn = nn.BatchNorm2d(128, affine=True)
        self.inception_4e_relu_3x3_reduce = nn.ReLU (inplace)
        self.inception_4e_3x3 = nn.Conv2d(128, 192, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))
        self.inception_4e_3x3_bn = nn.BatchNorm2d(192, affine=True)
        self.inception_4e_relu_3x3 = nn.ReLU (inplace)
        self.inception_4e_double_3x3_reduce = nn.Conv2d(608, 192, kernel_size=(1, 1), stride=(1, 1))
        self.inception_4e_double_3x3_reduce_bn = nn.BatchNorm2d(192, affine=True)
        self.inception_4e_relu_double_3x3_reduce = nn.ReLU (inplace)
        self.inception_4e_double_3x3_1 = nn.Conv2d(192, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_4e_double_3x3_1_bn = nn.BatchNorm2d(256, affine=True)
        self.inception_4e_relu_double_3x3_1 = nn.ReLU (inplace)
        self.inception_4e_double_3x3_2 = nn.Conv2d(256, 256, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))
        self.inception_4e_double_3x3_2_bn = nn.BatchNorm2d(256, affine=True)
        self.inception_4e_relu_double_3x3_2 = nn.ReLU (inplace)
        self.inception_4e_pool = nn.MaxPool2d ((3, 3), stride=(2, 2), dilation=(1, 1), ceil_mode=True)
        self.inception_5a_1x1 = nn.Conv2d(1056, 352, kernel_size=(1, 1), stride=(1, 1))
        self.inception_5a_1x1_bn = nn.BatchNorm2d(352, affine=True)
        self.inception_5a_relu_1x1 = nn.ReLU (inplace)
        self.inception_5a_3x3_reduce = nn.Conv2d(1056, 192, kernel_size=(1, 1), stride=(1, 1))
        self.inception_5a_3x3_reduce_bn = nn.BatchNorm2d(192, affine=True)
        self.inception_5a_relu_3x3_reduce = nn.ReLU (inplace)
        self.inception_5a_3x3 = nn.Conv2d(192, 320, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_5a_3x3_bn = nn.BatchNorm2d(320, affine=True)
        self.inception_5a_relu_3x3 = nn.ReLU (inplace)
        self.inception_5a_double_3x3_reduce = nn.Conv2d(1056, 160, kernel_size=(1, 1), stride=(1, 1))
        self.inception_5a_double_3x3_reduce_bn = nn.BatchNorm2d(160, affine=True)
        self.inception_5a_relu_double_3x3_reduce = nn.ReLU (inplace)
        self.inception_5a_double_3x3_1 = nn.Conv2d(160, 224, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_5a_double_3x3_1_bn = nn.BatchNorm2d(224, affine=True)
        self.inception_5a_relu_double_3x3_1 = nn.ReLU (inplace)
        self.inception_5a_double_3x3_2 = nn.Conv2d(224, 224, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_5a_double_3x3_2_bn = nn.BatchNorm2d(224, affine=True)
        self.inception_5a_relu_double_3x3_2 = nn.ReLU (inplace)
        self.inception_5a_pool = nn.AvgPool2d (3, stride=1, padding=1, ceil_mode=True, count_include_pad=True)
        self.inception_5a_pool_proj = nn.Conv2d(1056, 128, kernel_size=(1, 1), stride=(1, 1))
        self.inception_5a_pool_proj_bn = nn.BatchNorm2d(128, affine=True)
        self.inception_5a_relu_pool_proj = nn.ReLU (inplace)
        self.inception_5b_1x1 = nn.Conv2d(1024, 352, kernel_size=(1, 1), stride=(1, 1))
        self.inception_5b_1x1_bn = nn.BatchNorm2d(352, affine=True)
        self.inception_5b_relu_1x1 = nn.ReLU (inplace)
        self.inception_5b_3x3_reduce = nn.Conv2d(1024, 192, kernel_size=(1, 1), stride=(1, 1))
        self.inception_5b_3x3_reduce_bn = nn.BatchNorm2d(192, affine=True)
        self.inception_5b_relu_3x3_reduce = nn.ReLU (inplace)
        self.inception_5b_3x3 = nn.Conv2d(192, 320, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_5b_3x3_bn = nn.BatchNorm2d(320, affine=True)
        self.inception_5b_relu_3x3 = nn.ReLU (inplace)
        self.inception_5b_double_3x3_reduce = nn.Conv2d(1024, 192, kernel_size=(1, 1), stride=(1, 1))
        self.inception_5b_double_3x3_reduce_bn = nn.BatchNorm2d(192, affine=True)
        self.inception_5b_relu_double_3x3_reduce = nn.ReLU (inplace)
        self.inception_5b_double_3x3_1 = nn.Conv2d(192, 224, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_5b_double_3x3_1_bn = nn.BatchNorm2d(224, affine=True)
        self.inception_5b_relu_double_3x3_1 = nn.ReLU (inplace)
        self.inception_5b_double_3x3_2 = nn.Conv2d(224, 224, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.inception_5b_double_3x3_2_bn = nn.BatchNorm2d(224, affine=True)
        self.inception_5b_relu_double_3x3_2 = nn.ReLU (inplace)
        self.inception_5b_pool = nn.MaxPool2d ((3, 3), stride=(1, 1), padding=(1, 1), dilation=(1, 1), ceil_mode=True)
        self.inception_5b_pool_proj = nn.Conv2d(1024, 128, kernel_size=(1, 1), stride=(1, 1))
        self.inception_5b_pool_proj_bn = nn.BatchNorm2d(128, affine=True)
        self.inception_5b_relu_pool_proj = nn.ReLU (inplace)
        self.global_pool = nn.AvgPool2d (7, stride=1, padding=0, ceil_mode=True, count_include_pad=True)
        self.last_linear = nn.Linear (1024, num_classes)

        self.adavgpool = nn.AdaptiveAvgPool2d(1)
        self.admaxpool = nn.AdaptiveMaxPool2d(1)

    def features(self, input):
        conv1_7x7_s2_out = self.conv1_7x7_s2(input)
        conv1_7x7_s2_bn_out = self.conv1_7x7_s2_bn(conv1_7x7_s2_out)
        conv1_relu_7x7_out = self.conv1_relu_7x7(conv1_7x7_s2_bn_out)
        pool1_3x3_s2_out = self.pool1_3x3_s2(conv1_relu_7x7_out)
        conv2_3x3_reduce_out = self.conv2_3x3_reduce(pool1_3x3_s2_out)
        conv2_3x3_reduce_bn_out = self.conv2_3x3_reduce_bn(conv2_3x3_reduce_out)
        conv2_relu_3x3_reduce_out = self.conv2_relu_3x3_reduce(conv2_3x3_reduce_bn_out)
        conv2_3x3_out = self.conv2_3x3(conv2_relu_3x3_reduce_out)
        conv2_3x3_bn_out = self.conv2_3x3_bn(conv2_3x3_out)
        conv2_relu_3x3_out = self.conv2_relu_3x3(conv2_3x3_bn_out)
        pool2_3x3_s2_out = self.pool2_3x3_s2(conv2_relu_3x3_out)
        inception_3a_1x1_out = self.inception_3a_1x1(pool2_3x3_s2_out)
        inception_3a_1x1_bn_out = self.inception_3a_1x1_bn(inception_3a_1x1_out)
        inception_3a_relu_1x1_out = self.inception_3a_relu_1x1(inception_3a_1x1_bn_out)
        inception_3a_3x3_reduce_out = self.inception_3a_3x3_reduce(pool2_3x3_s2_out)
        inception_3a_3x3_reduce_bn_out = self.inception_3a_3x3_reduce_bn(inception_3a_3x3_reduce_out)
        inception_3a_relu_3x3_reduce_out = self.inception_3a_relu_3x3_reduce(inception_3a_3x3_reduce_bn_out)
        inception_3a_3x3_out = self.inception_3a_3x3(inception_3a_relu_3x3_reduce_out)
        inception_3a_3x3_bn_out = self.inception_3a_3x3_bn(inception_3a_3x3_out)
        inception_3a_relu_3x3_out = self.inception_3a_relu_3x3(inception_3a_3x3_bn_out)
        inception_3a_double_3x3_reduce_out = self.inception_3a_double_3x3_reduce(pool2_3x3_s2_out)
        inception_3a_double_3x3_reduce_bn_out = self.inception_3a_double_3x3_reduce_bn(inception_3a_double_3x3_reduce_out)
        inception_3a_relu_double_3x3_reduce_out = self.inception_3a_relu_double_3x3_reduce(inception_3a_double_3x3_reduce_bn_out)
        inception_3a_double_3x3_1_out = self.inception_3a_double_3x3_1(inception_3a_relu_double_3x3_reduce_out)
        inception_3a_double_3x3_1_bn_out = self.inception_3a_double_3x3_1_bn(inception_3a_double_3x3_1_out)
        inception_3a_relu_double_3x3_1_out = self.inception_3a_relu_double_3x3_1(inception_3a_double_3x3_1_bn_out)
        inception_3a_double_3x3_2_out = self.inception_3a_double_3x3_2(inception_3a_relu_double_3x3_1_out)
        inception_3a_double_3x3_2_bn_out = self.inception_3a_double_3x3_2_bn(inception_3a_double_3x3_2_out)
        inception_3a_relu_double_3x3_2_out = self.inception_3a_relu_double_3x3_2(inception_3a_double_3x3_2_bn_out)
        inception_3a_pool_out = self.inception_3a_pool(pool2_3x3_s2_out)
        inception_3a_pool_proj_out = self.inception_3a_pool_proj(inception_3a_pool_out)
        inception_3a_pool_proj_bn_out = self.inception_3a_pool_proj_bn(inception_3a_pool_proj_out)
        inception_3a_relu_pool_proj_out = self.inception_3a_relu_pool_proj(inception_3a_pool_proj_bn_out)
        inception_3a_output_out = torch.cat([inception_3a_relu_1x1_out,inception_3a_relu_3x3_out,inception_3a_relu_double_3x3_2_out ,inception_3a_relu_pool_proj_out], 1)
        inception_3b_1x1_out = self.inception_3b_1x1(inception_3a_output_out)
        inception_3b_1x1_bn_out = self.inception_3b_1x1_bn(inception_3b_1x1_out)
        inception_3b_relu_1x1_out = self.inception_3b_relu_1x1(inception_3b_1x1_bn_out)
        inception_3b_3x3_reduce_out = self.inception_3b_3x3_reduce(inception_3a_output_out)
        inception_3b_3x3_reduce_bn_out = self.inception_3b_3x3_reduce_bn(inception_3b_3x3_reduce_out)
        inception_3b_relu_3x3_reduce_out = self.inception_3b_relu_3x3_reduce(inception_3b_3x3_reduce_bn_out)
        inception_3b_3x3_out = self.inception_3b_3x3(inception_3b_relu_3x3_reduce_out)
        inception_3b_3x3_bn_out = self.inception_3b_3x3_bn(inception_3b_3x3_out)
        inception_3b_relu_3x3_out = self.inception_3b_relu_3x3(inception_3b_3x3_bn_out)
        inception_3b_double_3x3_reduce_out = self.inception_3b_double_3x3_reduce(inception_3a_output_out)
        inception_3b_double_3x3_reduce_bn_out = self.inception_3b_double_3x3_reduce_bn(inception_3b_double_3x3_reduce_out)
        inception_3b_relu_double_3x3_reduce_out = self.inception_3b_relu_double_3x3_reduce(inception_3b_double_3x3_reduce_bn_out)
        inception_3b_double_3x3_1_out = self.inception_3b_double_3x3_1(inception_3b_relu_double_3x3_reduce_out)
        inception_3b_double_3x3_1_bn_out = self.inception_3b_double_3x3_1_bn(inception_3b_double_3x3_1_out)
        inception_3b_relu_double_3x3_1_out = self.inception_3b_relu_double_3x3_1(inception_3b_double_3x3_1_bn_out)
        inception_3b_double_3x3_2_out = self.inception_3b_double_3x3_2(inception_3b_relu_double_3x3_1_out)
        inception_3b_double_3x3_2_bn_out = self.inception_3b_double_3x3_2_bn(inception_3b_double_3x3_2_out)
        inception_3b_relu_double_3x3_2_out = self.inception_3b_relu_double_3x3_2(inception_3b_double_3x3_2_bn_out)
        inception_3b_pool_out = self.inception_3b_pool(inception_3a_output_out)
        inception_3b_pool_proj_out = self.inception_3b_pool_proj(inception_3b_pool_out)
        inception_3b_pool_proj_bn_out = self.inception_3b_pool_proj_bn(inception_3b_pool_proj_out)
        inception_3b_relu_pool_proj_out = self.inception_3b_relu_pool_proj(inception_3b_pool_proj_bn_out)
        inception_3b_output_out = torch.cat([inception_3b_relu_1x1_out,inception_3b_relu_3x3_out,inception_3b_relu_double_3x3_2_out,inception_3b_relu_pool_proj_out], 1)
        inception_3c_3x3_reduce_out = self.inception_3c_3x3_reduce(inception_3b_output_out)
        inception_3c_3x3_reduce_bn_out = self.inception_3c_3x3_reduce_bn(inception_3c_3x3_reduce_out)
        inception_3c_relu_3x3_reduce_out = self.inception_3c_relu_3x3_reduce(inception_3c_3x3_reduce_bn_out)
        inception_3c_3x3_out = self.inception_3c_3x3(inception_3c_relu_3x3_reduce_out)
        inception_3c_3x3_bn_out = self.inception_3c_3x3_bn(inception_3c_3x3_out)
        inception_3c_relu_3x3_out = self.inception_3c_relu_3x3(inception_3c_3x3_bn_out)
        inception_3c_double_3x3_reduce_out = self.inception_3c_double_3x3_reduce(inception_3b_output_out)
        inception_3c_double_3x3_reduce_bn_out = self.inception_3c_double_3x3_reduce_bn(inception_3c_double_3x3_reduce_out)
        inception_3c_relu_double_3x3_reduce_out = self.inception_3c_relu_double_3x3_reduce(inception_3c_double_3x3_reduce_bn_out)
        inception_3c_double_3x3_1_out = self.inception_3c_double_3x3_1(inception_3c_relu_double_3x3_reduce_out)
        inception_3c_double_3x3_1_bn_out = self.inception_3c_double_3x3_1_bn(inception_3c_double_3x3_1_out)
        inception_3c_relu_double_3x3_1_out = self.inception_3c_relu_double_3x3_1(inception_3c_double_3x3_1_bn_out)
        inception_3c_double_3x3_2_out = self.inception_3c_double_3x3_2(inception_3c_relu_double_3x3_1_out)
        inception_3c_double_3x3_2_bn_out = self.inception_3c_double_3x3_2_bn(inception_3c_double_3x3_2_out)
        inception_3c_relu_double_3x3_2_out = self.inception_3c_relu_double_3x3_2(inception_3c_double_3x3_2_bn_out)
        inception_3c_pool_out = self.inception_3c_pool(inception_3b_output_out)
        inception_3c_output_out = torch.cat([inception_3c_relu_3x3_out,inception_3c_relu_double_3x3_2_out,inception_3c_pool_out], 1)
        inception_4a_1x1_out = self.inception_4a_1x1(inception_3c_output_out)
        inception_4a_1x1_bn_out = self.inception_4a_1x1_bn(inception_4a_1x1_out)
        inception_4a_relu_1x1_out = self.inception_4a_relu_1x1(inception_4a_1x1_bn_out)
        inception_4a_3x3_reduce_out = self.inception_4a_3x3_reduce(inception_3c_output_out)
        inception_4a_3x3_reduce_bn_out = self.inception_4a_3x3_reduce_bn(inception_4a_3x3_reduce_out)
        inception_4a_relu_3x3_reduce_out = self.inception_4a_relu_3x3_reduce(inception_4a_3x3_reduce_bn_out)
        inception_4a_3x3_out = self.inception_4a_3x3(inception_4a_relu_3x3_reduce_out)
        inception_4a_3x3_bn_out = self.inception_4a_3x3_bn(inception_4a_3x3_out)
        inception_4a_relu_3x3_out = self.inception_4a_relu_3x3(inception_4a_3x3_bn_out)
        inception_4a_double_3x3_reduce_out = self.inception_4a_double_3x3_reduce(inception_3c_output_out)
        inception_4a_double_3x3_reduce_bn_out = self.inception_4a_double_3x3_reduce_bn(inception_4a_double_3x3_reduce_out)
        inception_4a_relu_double_3x3_reduce_out = self.inception_4a_relu_double_3x3_reduce(inception_4a_double_3x3_reduce_bn_out)
        inception_4a_double_3x3_1_out = self.inception_4a_double_3x3_1(inception_4a_relu_double_3x3_reduce_out)
        inception_4a_double_3x3_1_bn_out = self.inception_4a_double_3x3_1_bn(inception_4a_double_3x3_1_out)
        inception_4a_relu_double_3x3_1_out = self.inception_4a_relu_double_3x3_1(inception_4a_double_3x3_1_bn_out)
        inception_4a_double_3x3_2_out = self.inception_4a_double_3x3_2(inception_4a_relu_double_3x3_1_out)
        inception_4a_double_3x3_2_bn_out = self.inception_4a_double_3x3_2_bn(inception_4a_double_3x3_2_out)
        inception_4a_relu_double_3x3_2_out = self.inception_4a_relu_double_3x3_2(inception_4a_double_3x3_2_bn_out)
        inception_4a_pool_out = self.inception_4a_pool(inception_3c_output_out)
        inception_4a_pool_proj_out = self.inception_4a_pool_proj(inception_4a_pool_out)
        inception_4a_pool_proj_bn_out = self.inception_4a_pool_proj_bn(inception_4a_pool_proj_out)
        inception_4a_relu_pool_proj_out = self.inception_4a_relu_pool_proj(inception_4a_pool_proj_bn_out)
        inception_4a_output_out = torch.cat([inception_4a_relu_1x1_out,inception_4a_relu_3x3_out,inception_4a_relu_double_3x3_2_out,inception_4a_relu_pool_proj_out], 1)
        inception_4b_1x1_out = self.inception_4b_1x1(inception_4a_output_out)
        inception_4b_1x1_bn_out = self.inception_4b_1x1_bn(inception_4b_1x1_out)
        inception_4b_relu_1x1_out = self.inception_4b_relu_1x1(inception_4b_1x1_bn_out)
        inception_4b_3x3_reduce_out = self.inception_4b_3x3_reduce(inception_4a_output_out)
        inception_4b_3x3_reduce_bn_out = self.inception_4b_3x3_reduce_bn(inception_4b_3x3_reduce_out)
        inception_4b_relu_3x3_reduce_out = self.inception_4b_relu_3x3_reduce(inception_4b_3x3_reduce_bn_out)
        inception_4b_3x3_out = self.inception_4b_3x3(inception_4b_relu_3x3_reduce_out)
        inception_4b_3x3_bn_out = self.inception_4b_3x3_bn(inception_4b_3x3_out)
        inception_4b_relu_3x3_out = self.inception_4b_relu_3x3(inception_4b_3x3_bn_out)
        inception_4b_double_3x3_reduce_out = self.inception_4b_double_3x3_reduce(inception_4a_output_out)
        inception_4b_double_3x3_reduce_bn_out = self.inception_4b_double_3x3_reduce_bn(inception_4b_double_3x3_reduce_out)
        inception_4b_relu_double_3x3_reduce_out = self.inception_4b_relu_double_3x3_reduce(inception_4b_double_3x3_reduce_bn_out)
        inception_4b_double_3x3_1_out = self.inception_4b_double_3x3_1(inception_4b_relu_double_3x3_reduce_out)
        inception_4b_double_3x3_1_bn_out = self.inception_4b_double_3x3_1_bn(inception_4b_double_3x3_1_out)
        inception_4b_relu_double_3x3_1_out = self.inception_4b_relu_double_3x3_1(inception_4b_double_3x3_1_bn_out)
        inception_4b_double_3x3_2_out = self.inception_4b_double_3x3_2(inception_4b_relu_double_3x3_1_out)
        inception_4b_double_3x3_2_bn_out = self.inception_4b_double_3x3_2_bn(inception_4b_double_3x3_2_out)
        inception_4b_relu_double_3x3_2_out = self.inception_4b_relu_double_3x3_2(inception_4b_double_3x3_2_bn_out)
        inception_4b_pool_out = self.inception_4b_pool(inception_4a_output_out)
        inception_4b_pool_proj_out = self.inception_4b_pool_proj(inception_4b_pool_out)
        inception_4b_pool_proj_bn_out = self.inception_4b_pool_proj_bn(inception_4b_pool_proj_out)
        inception_4b_relu_pool_proj_out = self.inception_4b_relu_pool_proj(inception_4b_pool_proj_bn_out)
        inception_4b_output_out = torch.cat([inception_4b_relu_1x1_out,inception_4b_relu_3x3_out,inception_4b_relu_double_3x3_2_out,inception_4b_relu_pool_proj_out], 1)
        inception_4c_1x1_out = self.inception_4c_1x1(inception_4b_output_out)
        inception_4c_1x1_bn_out = self.inception_4c_1x1_bn(inception_4c_1x1_out)
        inception_4c_relu_1x1_out = self.inception_4c_relu_1x1(inception_4c_1x1_bn_out)
        inception_4c_3x3_reduce_out = self.inception_4c_3x3_reduce(inception_4b_output_out)
        inception_4c_3x3_reduce_bn_out = self.inception_4c_3x3_reduce_bn(inception_4c_3x3_reduce_out)
        inception_4c_relu_3x3_reduce_out = self.inception_4c_relu_3x3_reduce(inception_4c_3x3_reduce_bn_out)
        inception_4c_3x3_out = self.inception_4c_3x3(inception_4c_relu_3x3_reduce_out)
        inception_4c_3x3_bn_out = self.inception_4c_3x3_bn(inception_4c_3x3_out)
        inception_4c_relu_3x3_out = self.inception_4c_relu_3x3(inception_4c_3x3_bn_out)
        inception_4c_double_3x3_reduce_out = self.inception_4c_double_3x3_reduce(inception_4b_output_out)
        inception_4c_double_3x3_reduce_bn_out = self.inception_4c_double_3x3_reduce_bn(inception_4c_double_3x3_reduce_out)
        inception_4c_relu_double_3x3_reduce_out = self.inception_4c_relu_double_3x3_reduce(inception_4c_double_3x3_reduce_bn_out)
        inception_4c_double_3x3_1_out = self.inception_4c_double_3x3_1(inception_4c_relu_double_3x3_reduce_out)
        inception_4c_double_3x3_1_bn_out = self.inception_4c_double_3x3_1_bn(inception_4c_double_3x3_1_out)
        inception_4c_relu_double_3x3_1_out = self.inception_4c_relu_double_3x3_1(inception_4c_double_3x3_1_bn_out)
        inception_4c_double_3x3_2_out = self.inception_4c_double_3x3_2(inception_4c_relu_double_3x3_1_out)
        inception_4c_double_3x3_2_bn_out = self.inception_4c_double_3x3_2_bn(inception_4c_double_3x3_2_out)
        inception_4c_relu_double_3x3_2_out = self.inception_4c_relu_double_3x3_2(inception_4c_double_3x3_2_bn_out)
        inception_4c_pool_out = self.inception_4c_pool(inception_4b_output_out)
        inception_4c_pool_proj_out = self.inception_4c_pool_proj(inception_4c_pool_out)
        inception_4c_pool_proj_bn_out = self.inception_4c_pool_proj_bn(inception_4c_pool_proj_out)
        inception_4c_relu_pool_proj_out = self.inception_4c_relu_pool_proj(inception_4c_pool_proj_bn_out)
        inception_4c_output_out = torch.cat([inception_4c_relu_1x1_out,inception_4c_relu_3x3_out,inception_4c_relu_double_3x3_2_out,inception_4c_relu_pool_proj_out], 1)
        inception_4d_1x1_out = self.inception_4d_1x1(inception_4c_output_out)
        inception_4d_1x1_bn_out = self.inception_4d_1x1_bn(inception_4d_1x1_out)
        inception_4d_relu_1x1_out = self.inception_4d_relu_1x1(inception_4d_1x1_bn_out)
        inception_4d_3x3_reduce_out = self.inception_4d_3x3_reduce(inception_4c_output_out)
        inception_4d_3x3_reduce_bn_out = self.inception_4d_3x3_reduce_bn(inception_4d_3x3_reduce_out)
        inception_4d_relu_3x3_reduce_out = self.inception_4d_relu_3x3_reduce(inception_4d_3x3_reduce_bn_out)
        inception_4d_3x3_out = self.inception_4d_3x3(inception_4d_relu_3x3_reduce_out)
        inception_4d_3x3_bn_out = self.inception_4d_3x3_bn(inception_4d_3x3_out)
        inception_4d_relu_3x3_out = self.inception_4d_relu_3x3(inception_4d_3x3_bn_out)
        inception_4d_double_3x3_reduce_out = self.inception_4d_double_3x3_reduce(inception_4c_output_out)
        inception_4d_double_3x3_reduce_bn_out = self.inception_4d_double_3x3_reduce_bn(inception_4d_double_3x3_reduce_out)
        inception_4d_relu_double_3x3_reduce_out = self.inception_4d_relu_double_3x3_reduce(inception_4d_double_3x3_reduce_bn_out)
        inception_4d_double_3x3_1_out = self.inception_4d_double_3x3_1(inception_4d_relu_double_3x3_reduce_out)
        inception_4d_double_3x3_1_bn_out = self.inception_4d_double_3x3_1_bn(inception_4d_double_3x3_1_out)
        inception_4d_relu_double_3x3_1_out = self.inception_4d_relu_double_3x3_1(inception_4d_double_3x3_1_bn_out)
        inception_4d_double_3x3_2_out = self.inception_4d_double_3x3_2(inception_4d_relu_double_3x3_1_out)
        inception_4d_double_3x3_2_bn_out = self.inception_4d_double_3x3_2_bn(inception_4d_double_3x3_2_out)
        inception_4d_relu_double_3x3_2_out = self.inception_4d_relu_double_3x3_2(inception_4d_double_3x3_2_bn_out)
        inception_4d_pool_out = self.inception_4d_pool(inception_4c_output_out)
        inception_4d_pool_proj_out = self.inception_4d_pool_proj(inception_4d_pool_out)
        inception_4d_pool_proj_bn_out = self.inception_4d_pool_proj_bn(inception_4d_pool_proj_out)
        inception_4d_relu_pool_proj_out = self.inception_4d_relu_pool_proj(inception_4d_pool_proj_bn_out)
        inception_4d_output_out = torch.cat([inception_4d_relu_1x1_out,inception_4d_relu_3x3_out,inception_4d_relu_double_3x3_2_out,inception_4d_relu_pool_proj_out], 1)
        inception_4e_3x3_reduce_out = self.inception_4e_3x3_reduce(inception_4d_output_out)
        inception_4e_3x3_reduce_bn_out = self.inception_4e_3x3_reduce_bn(inception_4e_3x3_reduce_out)
        inception_4e_relu_3x3_reduce_out = self.inception_4e_relu_3x3_reduce(inception_4e_3x3_reduce_bn_out)
        inception_4e_3x3_out = self.inception_4e_3x3(inception_4e_relu_3x3_reduce_out)
        inception_4e_3x3_bn_out = self.inception_4e_3x3_bn(inception_4e_3x3_out)
        inception_4e_relu_3x3_out = self.inception_4e_relu_3x3(inception_4e_3x3_bn_out)
        inception_4e_double_3x3_reduce_out = self.inception_4e_double_3x3_reduce(inception_4d_output_out)
        inception_4e_double_3x3_reduce_bn_out = self.inception_4e_double_3x3_reduce_bn(inception_4e_double_3x3_reduce_out)
        inception_4e_relu_double_3x3_reduce_out = self.inception_4e_relu_double_3x3_reduce(inception_4e_double_3x3_reduce_bn_out)
        inception_4e_double_3x3_1_out = self.inception_4e_double_3x3_1(inception_4e_relu_double_3x3_reduce_out)
        inception_4e_double_3x3_1_bn_out = self.inception_4e_double_3x3_1_bn(inception_4e_double_3x3_1_out)
        inception_4e_relu_double_3x3_1_out = self.inception_4e_relu_double_3x3_1(inception_4e_double_3x3_1_bn_out)
        inception_4e_double_3x3_2_out = self.inception_4e_double_3x3_2(inception_4e_relu_double_3x3_1_out)
        inception_4e_double_3x3_2_bn_out = self.inception_4e_double_3x3_2_bn(inception_4e_double_3x3_2_out)
        inception_4e_relu_double_3x3_2_out = self.inception_4e_relu_double_3x3_2(inception_4e_double_3x3_2_bn_out)
        inception_4e_pool_out = self.inception_4e_pool(inception_4d_output_out)
        inception_4e_output_out = torch.cat([inception_4e_relu_3x3_out,inception_4e_relu_double_3x3_2_out,inception_4e_pool_out], 1)
        inception_5a_1x1_out = self.inception_5a_1x1(inception_4e_output_out)
        inception_5a_1x1_bn_out = self.inception_5a_1x1_bn(inception_5a_1x1_out)
        inception_5a_relu_1x1_out = self.inception_5a_relu_1x1(inception_5a_1x1_bn_out)
        inception_5a_3x3_reduce_out = self.inception_5a_3x3_reduce(inception_4e_output_out)
        inception_5a_3x3_reduce_bn_out = self.inception_5a_3x3_reduce_bn(inception_5a_3x3_reduce_out)
        inception_5a_relu_3x3_reduce_out = self.inception_5a_relu_3x3_reduce(inception_5a_3x3_reduce_bn_out)
        inception_5a_3x3_out = self.inception_5a_3x3(inception_5a_relu_3x3_reduce_out)
        inception_5a_3x3_bn_out = self.inception_5a_3x3_bn(inception_5a_3x3_out)
        inception_5a_relu_3x3_out = self.inception_5a_relu_3x3(inception_5a_3x3_bn_out)
        inception_5a_double_3x3_reduce_out = self.inception_5a_double_3x3_reduce(inception_4e_output_out)
        inception_5a_double_3x3_reduce_bn_out = self.inception_5a_double_3x3_reduce_bn(inception_5a_double_3x3_reduce_out)
        inception_5a_relu_double_3x3_reduce_out = self.inception_5a_relu_double_3x3_reduce(inception_5a_double_3x3_reduce_bn_out)
        inception_5a_double_3x3_1_out = self.inception_5a_double_3x3_1(inception_5a_relu_double_3x3_reduce_out)
        inception_5a_double_3x3_1_bn_out = self.inception_5a_double_3x3_1_bn(inception_5a_double_3x3_1_out)
        inception_5a_relu_double_3x3_1_out = self.inception_5a_relu_double_3x3_1(inception_5a_double_3x3_1_bn_out)
        inception_5a_double_3x3_2_out = self.inception_5a_double_3x3_2(inception_5a_relu_double_3x3_1_out)
        inception_5a_double_3x3_2_bn_out = self.inception_5a_double_3x3_2_bn(inception_5a_double_3x3_2_out)
        inception_5a_relu_double_3x3_2_out = self.inception_5a_relu_double_3x3_2(inception_5a_double_3x3_2_bn_out)
        inception_5a_pool_out = self.inception_5a_pool(inception_4e_output_out)
        inception_5a_pool_proj_out = self.inception_5a_pool_proj(inception_5a_pool_out)
        inception_5a_pool_proj_bn_out = self.inception_5a_pool_proj_bn(inception_5a_pool_proj_out)
        inception_5a_relu_pool_proj_out = self.inception_5a_relu_pool_proj(inception_5a_pool_proj_bn_out)
        inception_5a_output_out = torch.cat([inception_5a_relu_1x1_out,inception_5a_relu_3x3_out,inception_5a_relu_double_3x3_2_out,inception_5a_relu_pool_proj_out], 1)
        inception_5b_1x1_out = self.inception_5b_1x1(inception_5a_output_out)
        inception_5b_1x1_bn_out = self.inception_5b_1x1_bn(inception_5b_1x1_out)
        inception_5b_relu_1x1_out = self.inception_5b_relu_1x1(inception_5b_1x1_bn_out)
        inception_5b_3x3_reduce_out = self.inception_5b_3x3_reduce(inception_5a_output_out)
        inception_5b_3x3_reduce_bn_out = self.inception_5b_3x3_reduce_bn(inception_5b_3x3_reduce_out)
        inception_5b_relu_3x3_reduce_out = self.inception_5b_relu_3x3_reduce(inception_5b_3x3_reduce_bn_out)
        inception_5b_3x3_out = self.inception_5b_3x3(inception_5b_relu_3x3_reduce_out)
        inception_5b_3x3_bn_out = self.inception_5b_3x3_bn(inception_5b_3x3_out)
        inception_5b_relu_3x3_out = self.inception_5b_relu_3x3(inception_5b_3x3_bn_out)
        inception_5b_double_3x3_reduce_out = self.inception_5b_double_3x3_reduce(inception_5a_output_out)
        inception_5b_double_3x3_reduce_bn_out = self.inception_5b_double_3x3_reduce_bn(inception_5b_double_3x3_reduce_out)
        inception_5b_relu_double_3x3_reduce_out = self.inception_5b_relu_double_3x3_reduce(inception_5b_double_3x3_reduce_bn_out)
        inception_5b_double_3x3_1_out = self.inception_5b_double_3x3_1(inception_5b_relu_double_3x3_reduce_out)
        inception_5b_double_3x3_1_bn_out = self.inception_5b_double_3x3_1_bn(inception_5b_double_3x3_1_out)
        inception_5b_relu_double_3x3_1_out = self.inception_5b_relu_double_3x3_1(inception_5b_double_3x3_1_bn_out)
        inception_5b_double_3x3_2_out = self.inception_5b_double_3x3_2(inception_5b_relu_double_3x3_1_out)
        inception_5b_double_3x3_2_bn_out = self.inception_5b_double_3x3_2_bn(inception_5b_double_3x3_2_out)
        inception_5b_relu_double_3x3_2_out = self.inception_5b_relu_double_3x3_2(inception_5b_double_3x3_2_bn_out)
        inception_5b_pool_out = self.inception_5b_pool(inception_5a_output_out)
        inception_5b_pool_proj_out = self.inception_5b_pool_proj(inception_5b_pool_out)
        inception_5b_pool_proj_bn_out = self.inception_5b_pool_proj_bn(inception_5b_pool_proj_out)
        inception_5b_relu_pool_proj_out = self.inception_5b_relu_pool_proj(inception_5b_pool_proj_bn_out)
        inception_5b_output_out = torch.cat([inception_5b_relu_1x1_out,inception_5b_relu_3x3_out,inception_5b_relu_double_3x3_2_out,inception_5b_relu_pool_proj_out], 1)
        return inception_5b_output_out
    
    def forward(self, input):
        feat = self.features(input)
        ap_feat = self.adavgpool(feat)
        mp_feat = self.admaxpool(feat)
        output = ap_feat + mp_feat
        output = output.view(output.size(0), -1)
        return output

def bninception(num_classes=1000, pretrained='imagenet',bn_freeze=True,freeze_all=False):
    r"""BNInception model architecture from <https://arxiv.org/pdf/1502.03167.pdf>`_ paper.
    """
    model = BNInception(num_classes=1000)
    if pretrained is not None:
        settings = pretrained_settings['bninception'][pretrained]
        assert num_classes == settings['num_classes'], \
            "num_classes should be {}, but is {}".format(settings['num_classes'], num_classes)
        model.load_state_dict(model_zoo.load_url(settings['url']))
        model.input_space = settings['input_space']
        model.input_size = settings['input_size']
        model.input_range = settings['input_range']
        model.mean = settings['mean']
        model.std = settings['std']
    
    # freeze all parameters
    if freeze_all:
        for param in model.parameters():
            param.requires_grad = False
    
    # bn freeze
    if bn_freeze:
        for m in model.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()
                m.weight.requires_grad_(False)
                m.bias.requires_grad_(False)
    
    return model


if __name__ == '__main__':

    model = bninception()
    data = torch.randn(32, 3, 227, 227)
    output = model(data)
    pass



import logging
import numpy as np 

import torch.distributed as dist 
import torch

try: 
    import faiss
except ModuleNotFoundError:
    logging.warning("Faiss Package Not Found!")

"""
(Faiss) to multi gpu devices
"""

def index_cpu_to_gpu_multiple(index, resources=None, co=None, gpu_ids=None):
    assert isinstance(gpu_ids, list)
    if resources is None:
        resources = [faiss.StandardGpuResources() for i in gpu_ids]
    v_res = faiss.GpuResourcesVector()
    v_dev = faiss.IntVector()
    if gpu_ids is None:
        gpu_ids = list(range(len(resources)))
    for i, res in zip(gpu_ids, resources):
        v_dev.push_back(i)
        v_res.push_back(res)
    index = faiss.index_cpu_to_gpu_multiple(v_res, v_dev, index, co)
    index.referenced_objects = resources
    return index

"""
Some random number generator
"""
def multi_center_generator(sample_num_per_class, class_num, dim=2, scale=0.1):
    loc = np.random.rand(class_num, dim)
    scale = scale * np.ones((class_num, dim))
    data = []
    for i in range(class_num):
        data.append(np.random.normal(loc=loc[i, :], scale=scale[i, :], size=(sample_num_per_class, dim)))
    data = np.vstack(data).astype('float32')
    label = np.arange(class_num).reshape(-1, 1)
    label = np.tile(label, (1, sample_num_per_class)).reshape(-1).astype('float32')
    return data, label

"""
Distributed utilities
"""
def distributed_gather_objects(*objects_list, rank=None, world_size=None):
    gathered_objects_list = []
    world_size = (
        dist.get_world_size()
        if world_size is None
        else world_size
    )
    rank = (
        dist.get_rank()
        if rank is None
        else rank
    )
    for objects_item in list(objects_list):
        curr_gather_list = [
            torch.zeros_like(objects_item)
            for _ in range(world_size)
        ]
        dist.all_gather(
            tensor_list=curr_gather_list,
            tensor=objects_item.contiguous()
        )
        curr_gather_list[rank] = objects_item
        curr_gather = torch.cat(curr_gather_list)
        gathered_objects_list.append(curr_gather)
    return tuple(gathered_objects_list)


import torch
from ..base_loss import BaseLoss

class ProxyLoss(BaseLoss):
    """
    paper: `No Fuss Distance Metric Learning Using Proxies <https://openaccess.thecvf.com/content_iccv_2017/html/Movshovitz-Attias_No_Fuss_Distance_ICCV_2017_paper.html>`_
    """
    def __init__(
        self,
        gamma,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.gamma = gamma
    
    def required_metric(self):
        return ["euclid"]
    
    def compute_loss(
        self,
        metric_mat,
        row_labels,
        col_labels,
        is_same_source=False,
        *args,
        **kwargs
    ) -> torch.Tensor:
        dtype, device = metric_mat.dtype, metric_mat.device
        pos_mask = (row_labels == col_labels).type(dtype).to(device)
        exp = torch.nn.functional.softmax(
            self.gamma * metric_mat, dim=-1
        )
        exp = torch.sum(
            exp * pos_mask, dim=-1
        )
        exp = exp[torch.where(exp)[0]]
        loss = torch.mean( - torch.log(exp))
        return loss


import numpy as np
import torch

from .base_selector import BaseSelector
from ...config.setting.core_setting import (
    INDICES_TUPLE,
    INDICES_FLAG
)

class DensePairSelector(BaseSelector):  
    """
    Select all pairs.
    """  
    def forward(
        self, 
        metric_mat, 
        row_labels, 
        col_labels, 
        is_same_source=False
    ) -> tuple:
        """
        Select all pairs.
        """
        pos_mask = row_labels == col_labels
        neg_mask = ~pos_mask
        if is_same_source:
            pos_mask.fill_diagonal_(False)
        pos_pairs = torch.stack(torch.where(pos_mask), dim=1)
        neg_pairs = torch.stack(torch.where(neg_mask), dim=1)
        tuples = torch.cat((pos_pairs, neg_pairs), dim=0) 
        pos_flags = torch.ones(pos_pairs.shape[0],1)
        neg_flags = torch.zeros(neg_pairs.shape[0],1)
        flags = torch.cat((pos_flags, neg_flags), dim=0).byte()
        weight = None
        indices_tuple = {INDICES_TUPLE:tuples, INDICES_FLAG:flags}
        return metric_mat, row_labels, col_labels, is_same_source, indices_tuple, weight


import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np 
from ..base_collector import BaseCollector

class HDMLCollector(BaseCollector):
    """
    Use variational autoencoder to decompose intra-class invariance and intra-class variance.

    Paper: `Hardness-Aware Deep Metric Learning <https://openaccess.thecvf.com/content_CVPR_2019/html/Zheng_Hardness-Aware_Deep_Metric_Learning_CVPR_2019_paper.html>`_

    Four types of loss: 
    (loss_avg = loss_m,
    loss_gen = loss_recon + loss_soft)
    
    1. loss_recon
    2. loss_soft
    3. loss_syn
    4. loss_m

    Args:
        generator (torch.nn.Module): multi-layer perceptron
        embedder (torch.nn.Module): multi-layer perceptron
        classifier (torch.nn.Module): multi-layer perceptrons
        alpha (float): 90.0 (NPairLoss) or 7.0 (TripletLoss)
        beta (float): 1.0e4
        coef_lambda (float): 0.5
        soft_weight (float): 1.0e4
        d_plus_scheme (str): default: ``positive_distance``
        d_plus (float): Constant or positive pair distance. default: 0.5
    """
    def __init__(
        self,
        generator,
        embedder,
        classifier,
        alpha=90.0,
        beta=1.0e4,
        coef_lambda=0.5,
        soft_weight=1.0e4,
        d_plus_scheme="positive_distance",
        d_plus=0.5,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.alpha = alpha
        self.beta = beta
        self.coef_lambda = coef_lambda
        self.soft_weight = soft_weight
        self.d_plus_scheme = d_plus_scheme # TODO: 
        self.d_plus = d_plus

        self.generator = generator
        self.embedder = embedder
        self.classifier = classifier

        self.lambda_0 = None
        self.loss_avg = None # adjust the hardness
        self.loss_gen = None # adjust the 
    
    
    def update(self, trainer):
        """
        In HDML paper, an adaptive weighting method is proposed. Therefore, before each epoch ``loss_avg`` and ``loss_gen`` must be updated from outside ``trainer``

        :math:`loss_{avg} = loss_m`

        :math:`loss_{gen} = loss_{recon} + loss_{soft}`
        """
        loss_handler = trainer.loss_handler
        self.loss_avg = float(loss_handler.get("loss_m", self._default_loss_value, is_avg_value=True))
        self.loss_gen = float(
            loss_handler.get("loss_recon", self._default_loss_value, is_avg_value=True) + 
            loss_handler.get("loss_soft", self._default_loss_value, is_avg_value=True)
        )
        if not torch.is_tensor(self.loss_avg):
            self.loss_avg = torch.tensor(self.loss_avg)
        if not torch.is_tensor(self.loss_gen):
            self.loss_gen = torch.tensor(self.loss_gen)
    
    def construct_neg_embedding_hat(self, pos_embedding, neg_embedding):
        # compute lambda_0
        np_dist = F.pairwise_distance(pos_embedding, neg_embedding, p=2)
        d_plus = torch.ones_like(np_dist, device=np_dist.device) * self.d_plus
        self.lambda_loss = torch.exp( - self.alpha / self.loss_avg)
        self.lambda_0 = (
            self.lambda_loss + ( 1 - self.lambda_loss ) 
            * d_plus / np_dist
        ).unsqueeze(-1).detach()
        self.lambda_0[np_dist <= d_plus] = 1

        # construct embedding_hat
        neg_embedding_hat = (
            pos_embedding + 
            self.lambda_0 * (
                neg_embedding - pos_embedding
            )
        )
        
        return pos_embedding, neg_embedding_hat

    def generate_feature(self, embeddings):
        # generate features
        return self.generator(embeddings)
    
    def forward(
        self, 
        data, 
        embeddings, 
        features,
        labels
    ) -> tuple:
        """
        Define four kinds of losses.

        :math:`loss_{total} = w_{recon} \\times loss_{recon} + w_{soft} \\times loss_{soft} + w_m \\times loss_m + w_{syn} \\times loss_{syn}`

        :math:`loss_{recon} = mean(|f_{pos} - f_{pos-recon}|^2_2)`

        :math:`loss_{soft} = CrossEntropy(Prob_{recon}, Labels_{recon})`

        :math:`loss_m = loss_{metric}(matrix_{m})`

        :math:`loss_syn = loss_{metric}(matrix_{syn})`
        """
        batch_size = embeddings.size(0)
        # sample pos_idx and neg_idx
        neg_mask = (labels.unsqueeze(1) != labels.unsqueeze(0)).byte()
        # pos_idx, neg_idx = torch.where(neg_mask)
        pos_idx, neg_idx = [], []
        for i in range(batch_size):
            neg_list = torch.where(neg_mask[i])[0]
            if len(neg_list) > 0:
                pos_idx.append(i)
                neg_idx.append(np.random.choice(neg_list.cpu().numpy()))
        pos_idx = torch.tensor(pos_idx)
        neg_idx = torch.tensor(neg_idx)

        # get pos and neg pairs
        pos_embedding, neg_embedding = embeddings[pos_idx], embeddings[neg_idx]
        pos_labels, neg_labels = labels[pos_idx], labels[neg_idx]

        # construct hard negative embeddings
        pos_embedding, neg_embedding = self.construct_neg_embedding_hat(
            pos_embedding,
            neg_embedding
        )

        # generate hard negative features
        pos_recon_features = self.generate_feature(pos_embedding)
        neg_recon_features = self.generate_feature(neg_embedding)

        # compute reconstruction loss (loss_recon)
        loss_recon = torch.mean(torch.sum((features[pos_idx] - pos_recon_features).pow(2), dim=-1))
        weight_recon = self.coef_lambda

        # compute loss_m
        metric_mat_m = self.metric(embeddings, embeddings)
        row_labels_m = labels.unsqueeze(1)
        col_labels_m = labels.unsqueeze(0)
        is_same_source = True

        # compute loss_syn
        recon_features = torch.cat([pos_recon_features, neg_recon_features], dim=0)
        recon_labels = torch.cat([pos_labels, neg_labels], dim=0)
        metric_mat_syn = self.metric(recon_features, recon_features)
        row_labels_syn = recon_labels.unsqueeze(1)
        col_labels_syn = recon_labels.unsqueeze(0)

        # compute loss_soft
        recon_prob = self.classifier(recon_features)
        loss_soft = F.cross_entropy(recon_prob, recon_labels, reduction="mean") * self.coef_lambda
        weight_soft = self.coef_lambda * self.soft_weight
        
        # weight of loss_m and loss_syn
        weight_m = torch.exp( - self.beta / self.loss_gen)
        weight_syn = 1 - weight_m

        return (
            metric_mat_m,
            row_labels_m,
            col_labels_m,
            metric_mat_syn,
            row_labels_syn,
            col_labels_syn,
            is_same_source,
            loss_recon,
            loss_soft,
            weight_recon,
            weight_soft,
            weight_m,
            weight_syn
        )
    


from .base_creator import BaseCreator
from ....core import datasets

class datasetsCreator(BaseCreator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    
    def prepare_packages(self):
        self.package = [datasets]


from .base_creator import BaseCreator
from ....core.samplers import collate_fn

class collatefnsCreator(BaseCreator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    
    def prepare_packages(self):
        self.package = [collate_fn]



from .base_creator import BaseCreator

class metricsCreator(BaseCreator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def prepare_packages(self):
        from ....core import metrics
        self.package = [metrics]


import torch
from collections import defaultdict

from . import utils

class Storage:    
    def __init__(self, wrapper_params):
        self.indices_dict = {}
        self.wrapper_params = wrapper_params

    def get(self, item_name):
        return getattr(self, item_name)
    
    def get_list(self, item_name_list):
        return [
            getattr(self, item_name)
            for item_name in item_name_list
        ]
    
    def get_dict(self, item_name_list):
        return {
            item_name: getattr(self, item_name, None) # NOTE: if not exist, return **None**
            for item_name in item_name_list
        }
    
    def filter_dict(self, item_name_list, total_dict):
        output_dict = {}
        for item_name in item_name_list:
            item_value = total_dict.get(item_name, None)
            if item_value is None:
                item_value = getattr(self, item_name, None)
            output_dict[item_name] = item_value
        return output_dict

    def add(self, item_name, value):
        setattr(self, item_name, value)
        
    def add_dict(self, item_dict: dict):
        for k in item_dict.keys():
            module, name, group = k.split("/")
            if self.indices_dict.get(module, None):
                if self.indices_dict[module].get(name, None):
                    if self.indices_dict[module][name].get(group, None):
                        self.indices_dict[module][name][group].update(item_dict[k])
                    else:
                        self.indices_dict[module][name][group] = item_dict[k]
                else:
                    self.indices_dict[module][name] = {
                        group: item_dict[k]
                    }
            else:
                self.indices_dict[module] = {
                    name: {group: item_dict[k]}
                }

    def tensors_to_device(self, item_name_list, device):
        for k in item_name_list:
            setattr(
                self,
                k,
                getattr(self, k).to(device)
            )
    
    def output_wrapper(self, sub_value_tuple, sub_k, sub_v):
        sub_item_output_dict = {
            wrapper_k: sub_value_tuple[wrapper_idx]
            for wrapper_k, wrapper_idx in sub_v.items()
        }
        module_type, name, group = sub_k.split('/')
        sub_k = '/'.join([module_type, name, group])
        return sub_item_output_dict, sub_k

    def update(self, modules, cur_module=None):
        """
        For tuple input: wrapper will convert tuple into dict
        For dict input: wrapper will convert dict into dict recursively
        """
        value_dict = defaultdict(dict)

        cur_module_dict = self.indices_dict.get(cur_module, {})
        cur_wrapper_dict = self.wrapper_params.get(cur_module, {})
        if cur_module_dict is not None:
            dict_len = len(cur_module_dict)
            for k, v in cur_module_dict.items():
                # get module
                if not isinstance(modules, dict):
                    module = modules
                    assert dict_len == 1 # if not dict, dict_len must be ONE!
                else:
                    if len(modules) == 1:
                        module = utils.get_first_dict_value(modules)
                    else:
                        module = modules[k]
                
                # pass parameters
                sub_wrapper_dict = cur_wrapper_dict[k]
                for group_key, params in v.items():
                    # get output tuple
                    sub_value_tuple = module( 
                        **self.filter_dict(
                            sub_wrapper_dict["input"],
                            total_dict=params
                        )
                    )
                    # check the type of output tuple and wrap the output
                    if isinstance(sub_value_tuple, dict):
                        for tuple_k, tuple_v in sub_value_tuple.items():
                            for sub_k, sub_v in sub_wrapper_dict["map"].items():
                                sub_item_output_dict, new_sub_k = self.output_wrapper(tuple_v, sub_k, sub_v)
                                value_dict[new_sub_k+"{}".format(tuple_k)+group_key].update(sub_item_output_dict)
                    else:
                        if not isinstance(sub_value_tuple, tuple):
                            sub_value_tuple = (sub_value_tuple,)
                        # output wrapper
                        for sub_k, sub_v in sub_wrapper_dict["map"].items():
                            sub_item_output_dict, new_sub_k = self.output_wrapper(sub_value_tuple, sub_k, sub_v)
                            value_dict[new_sub_k+group_key].update(sub_item_output_dict)

        # add to storage
        self.add_dict(
            value_dict
        )

        return value_dict
    
    def return_loss_dict(self):
        return self.indices_dict["FINISH"]

    def __repr__(self):
        output_str = "{}\n".format(self.__class__.__name__)
        for k, v in self.__dict__.items():
            output_str += "{}:\n{}\n".format(
                k, str(v)
            )
        return output_str

if __name__ == '__main__':
    test = Storage()
    device = torch.device("cuda:0")
    test.d0 = torch.randn(3, 4)
    test.d1 = torch.randn(3, 4)
    test.tensors_to_device(["d0", "d1"], device)

    pass



import torch
import torch.nn.functional as F 
import torch.distributions as distributions

from ..base_collector import BaseCollector

class DVMLCollector(BaseCollector):
    """
    Paper: `Deep Variational Metric Learning <https://openaccess.thecvf.com/content_ECCV_2018/html/Xudong_Lin_Deep_Variational_Metric_ECCV_2018_paper.html>`_

    Four losses:

    1. loss_kl: KL divergence between learned distribution and isotropic multivariate Gaussian
    2. loss_recon: reconstruction loss of original images and images generated by the decoder
    3. metric learning loss of learned intra-class invariance
    4. metric learning loss of the combination of sampled intra-class variance and learned intra-class invariance

    Default parameters recommended in the paper:

    1. lr = 0.0001
    2. T = 20 (for sample generation)
    3. batch_size = 128 (pair-based) or 120 (triplet-based)

    There are two phases during training:

    1. first phase: cut off the back-propagation of the gradients from the decoder network: 
    ``lambda_1`` = 1,
    ``lambda_2`` = 1,
    ``lambda_3`` = 0.1,
    ``lambda_4`` = 1,
    
    2. second phase: release the constraint:
    ``lambda_1`` = 0.8,
    ``lambda_2`` = 1,
    ``lambda_3`` = 0.2,
    ``lambda_4`` = 0.8,

    Args:
        embedder_mean (torch.nn.Module): multi-layer perceptron
        embedder_std (torch.nn.Module): multi-layer perceptron
        decoder (torch.nn.Module): multi-layer perceptron
        T (int): default: 20
        phase (int): 1 for ``first phase`` and 2 for ``second phase``
        lambda_1 (int): first phase: 1; second phase: 0.8
        lambda_2 (int): first phase: 1; second phase: 1
        lambda_3 (int): first phase: 0.1; second phase: 0.2
        lambda_4 (int): first phase: 1; second phase: 0.8
    """
    def __init__(
        self,
        embedder_mean,
        embedder_std,
        decoder,
        T=20,
        phase=1,
        lambda_1=None,
        lambda_2=None,
        lambda_3=None,
        lambda_4=None,
        *args,
        **kwargs
    ):
        super(DVMLCollector, self).__init__(*args, **kwargs)
        self.embedder_mean = embedder_mean
        self.embedder_std = embedder_std
        self.decoder = decoder

        self.T = T
        self.phase = phase
        self.lambda_1 = lambda_1
        self.lambda_2 = lambda_2
        self.lambda_3 = lambda_3
        self.lambda_4 = lambda_4

        self._initiate_default_lambda()
    
    def _initiate_default_lambda(self):
        """
        Set parameters ``lambda_x``
        """
        if self.phase == 1:
            self.lambda_1 = 1 if self.lambda_1 is None else self.lambda_1
            self.lambda_2 = 1 if self.lambda_2 is None else self.lambda_2
            self.lambda_3 = 0.1 if self.lambda_3 is None else self.lambda_3
            self.lambda_4 = 1 if self.lambda_4 is None else self.lambda_4
        elif self.phase == 2:
            self.lambda_1 = 1 if self.lambda_1 is None else self.lambda_1
            self.lambda_2 = 1 if self.lambda_2 is None else self.lambda_2
            self.lambda_3 = 0.1 if self.lambda_3 is None else self.lambda_3
            self.lambda_4 = 1 if self.lambda_4 is None else self.lambda_4
        else:
            raise KeyError("parameter 'phase' must be 1 or 2!")
    
    def update(self, trainer):
        pass

    def forward(
        self,
        data,
        embeddings,
        features,
        labels
    ) -> tuple:
        """
        Four losses should be computed in function ``collect``: 

        :math:`loss_{total} = \lambda_1 \\times loss_{kl} + \lambda_2 \\times loss_{recon} + \lambda_3 \\times loss_{syn} + \lambda_4 \\times loss_{invariant}`

        :math:`loss_{kl} = KL(p_{dist}, q_{dist})`

        :math:`loss_{recon} = mean(|f_{decode} - f_{ori}|^2_2)`

        :math:`loss_{syn} = loss_{metric}(matrix_{syn}, labels)`

        :math:`loss_{invariant} = loss_{metric}(matrix_{inv}, labels)`
        """
        embedding_size = embeddings.size(1)
        feature_size = features.size(1)
        is_same_source = True
        # compute mu and std
        mu = self.embedder_mean(features)
        log_var = self.embedder_std(features)
        std = torch.exp(log_var * 0.5)
        q_distributions = distributions.Normal(mu, std)
        p_distributions = distributions.Normal(0, 1)

        # for discriminate intra-class invariant features: loss_invariant
        metric_mat_inv = self.metric(embeddings, embeddings)
        row_labels = labels.unsqueeze(1)
        col_labels = labels.unsqueeze(0)

        # loss_kl
        loss_kl = distributions.kl_divergence(q_distributions, p_distributions).sum(-1).mean()

        # two phase
        if self.phase == 1:
            embeddings_var = q_distributions.sample()
            embeddings_syn = embeddings + embeddings_var
            features_decode = self.decoder(embeddings_syn.detach())
        elif self.phase == 2:
            embeddings_var = q_distributions.sample((self.T,))
            embeddings_syn = embeddings.unsqueeze(0) + embeddings_var
            features_decode = self.decoder(
                embeddings_syn.view(-1, embedding_size)
            ).view(self.T, -1, feature_size)
            # use the first synthetic 
            embeddings_syn = embeddings_syn[0]
        
        # loss_recon
        loss_recon = F.mse_loss(features_decode, features)

        # for discriminate synthesized features
        metric_mat_syn = self.metric(embeddings_syn, embeddings_syn)

        return (
            metric_mat_inv,
            row_labels,
            col_labels,
            is_same_source,
            metric_mat_syn,
            loss_kl,
            loss_recon,
            self.lambda_1,
            self.lambda_2,
            self.lambda_3,
            self.lambda_4
        )




from .contrastive_loss import ContrastiveLoss
from .margin_loss import MarginLoss
from .triplet_loss import TripletLoss
from .angular_loss import AngularLoss
from .circle_loss import CircleLoss
from .fast_ap_loss import FastAPLoss
from .lifted_structure_loss import LiftedStructureLoss
from .multi_similarity_loss import MultiSimilarityLoss
from .n_pair_loss import NPairLoss
from .signal_to_noise_ratio_loss import SignalToNoiseRatioLoss
from .pos_pair_loss import PosPairLoss


import torch
import torch.nn as nn 

from ..modules import WithRecorder

class Identity(WithRecorder):
    """
    Do nothing.
    """
    def __init__(self):
        super(Identity, self).__init__()
    
    def forward(self, features):
        return features


"""
List
+++++++++++++++++++++++++++++++++++

1. `CUB200 <http://www.vision.caltech.edu/visipedia/CUB-200.html>`_
2. `Cars196 <https://ai.stanford.edu/~jkrause/cars/car_dataset.html>`_
3. `Stanford Online Products <https://cvgl.stanford.edu/projects/lifted_struct/>`_
4. `ImageNet <http://www.image-net.org/>`_

Directory format
+++++++++++++++++++++++++++++++++++
.. code:: bash

    (dataset_root)
        - train
            - class0
                pic1.jpg
                ...
            ...
        - test
            ...
"""

from .cub200 import CUB200
from .cars196 import Cars196
from .online_products import OnlineProducts
from .imagenet import ImageNet
from .mini_imagenet import MiniImageNet